"""
The ProductDisplayType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import Union, List

from sarpy.io.xml.base import Serializable, SerializableArray, ParametersCollection
from sarpy.io.xml.descriptors import SerializableDescriptor, SerializableListDescriptor, \
    IntegerDescriptor, FloatDescriptor, StringDescriptor, StringEnumDescriptor, \
    ParametersDescriptor, SerializableArrayDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import FilterType, NewLookupTableType


#########
# NonInteractiveProcessing

class BandLUTType(NewLookupTableType):
    _fields = ('Predefined', 'Custom', 'k')
    _required = ('k', )
    _set_as_attribute = ('k', )
    # Descriptor
    k = IntegerDescriptor(
        'k', _required, strict=DEFAULT_STRICT, bounds=(1, 2**32), default_value=1,
        docstring='The array index.')

    def __init__(self, Predefined=None, Custom=None, k=None, **kwargs):
        """

        Parameters
        ----------
        Predefined : sarpy.io.product.sidd2_elements.blocks.PredefinedLookupType
        Custom : sarpy.io.product.sidd2_elements.blocks.CustomLookupType
        k : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.k = k
        super(BandLUTType, self).__init__(Predefined=Predefined, Custom=Custom, **kwargs)


class BandLUTArray(SerializableArray):
    _set_size = False
    _set_index = True
    _index_var_name = 'k'


class BandEqualizationType(Serializable):
    """

    """
    _fields = ('Algorithm', 'BandLUTs')
    _required = ('Algorithm', 'BandLUTs')
    _collections_tags = {'BandLUTs': {'array': True, 'child_tag': 'BandLUT'}}
    # Descriptor
    Algorithm = StringEnumDescriptor(
        'Algorithm', ('1DLUT', ), _required, strict=DEFAULT_STRICT, default_value='1DLUT',
        docstring='The algorithm type.')  # type: str
    BandLUTs = SerializableArrayDescriptor(
        'BandLUTs', BandLUTType, _collections_tags, _required, strict=DEFAULT_STRICT, array_extension=BandLUTArray,
        docstring='')  # type: Union[BandLUTArray, List[BandLUTType]]

    def __init__(self, Algorithm='1DLUT', BandLUTs=None, **kwargs):
        """

        Parameters
        ----------
        Algorithm : str
            `1DLUT` is currently the only allowed value.
        BandLUTs : BandLUTArray|List[BandLUTType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Algorithm = Algorithm
        self.BandLUTs = BandLUTs
        super(BandEqualizationType, self).__init__(**kwargs)


class ProductGenerationOptionsType(Serializable):
    """

    """
    _fields = ('BandEqualization', 'ModularTransferFunctionRestoration', 'DataRemapping', 'AsymmetricPixelCorrection')
    _required = ('DataRemapping', )
    # Descriptor
    BandEqualization = SerializableDescriptor(
        'BandEqualization', BandEqualizationType, _required, strict=DEFAULT_STRICT,
        docstring='Band equalization ensures that real-world neutral colors have equal digital count values '
                  '(i.e. are represented as neutral colors) across the dynamic range '
                  'of the imaged scene.')  # type: BandEqualizationType
    ModularTransferFunctionRestoration = SerializableDescriptor(
        'ModularTransferFunctionRestoration', FilterType, _required, strict=DEFAULT_STRICT,
        docstring=r'If present, the filter must not exceed :math:`15 \times 15`.')  # type: FilterType
    DataRemapping = SerializableDescriptor(
        'DataRemapping', NewLookupTableType, _required, strict=DEFAULT_STRICT,
        docstring='Data remapping refers to the specific need to convert the data of incoming, expanded or '
                  'uncompressed image band data to non-mapped image data.')  # type: NewLookupTableType
    AsymmetricPixelCorrection = SerializableDescriptor(
        'AsymmetricPixelCorrection', FilterType, _required, strict=DEFAULT_STRICT,
        docstring='The asymmetric pixel correction.')  # type: FilterType

    def __init__(self, BandEqualization=None, ModularTransferFunctionRestoration=None, DataRemapping=None, **kwargs):
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.BandEqualization = BandEqualization
        self.ModularTransferFunctionRestoration = ModularTransferFunctionRestoration
        self.DataRemapping = DataRemapping
        super(ProductGenerationOptionsType, self).__init__(**kwargs)


class RRDSType(Serializable):
    """
    RRDS type.
    """

    _fields = ('DownsamplingMethod', 'AntiAlias', 'Interpolation')
    _required = ('DownsamplingMethod', )
    # Descriptor
    DownsamplingMethod = StringEnumDescriptor(
        'DownsamplingMethod',
        ('DECIMATE', 'MAX PIXEL', 'AVERAGE', 'NEAREST NEIGHBOR', 'BILINEAR', 'LAGRANGE'),
        _required, strict=DEFAULT_STRICT,
        docstring='Algorithm used to perform RRDS downsampling')  # type: str
    AntiAlias = SerializableDescriptor(
        'AntiAlias', FilterType, _required, strict=DEFAULT_STRICT,
        docstring='The anti-aliasing filter. Should only be included if '
                  '`DownsamplingMethod= "DECIMATE"`')  # type: FilterType
    Interpolation = SerializableDescriptor(
        'Interpolation', FilterType, _required, strict=DEFAULT_STRICT,
        docstring='The interpolation filter. Should only be included if '
                  '`DownsamplingMethod= "DECIMATE"`')  # type: FilterType

    def __init__(self, DownsamplingMethod=None, AntiAlias=None, Interpolation=None, **kwargs):
        """

        Parameters
        ----------
        DownsamplingMethod : str
        AntiAlias : None|FilterType
        Interpolation : None|FilterType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.DownsamplingMethod = DownsamplingMethod
        self.AntiAlias = AntiAlias
        self.Interpolation = Interpolation
        super(RRDSType, self).__init__(**kwargs)


class NonInteractiveProcessingType(Serializable):
    """
    The non-interactive processing information.
    """

    _fields = ('ProductGenerationOptions', 'RRDS', 'band')
    _required = ('ProductGenerationOptions', 'RRDS', 'band')
    _set_as_attribute = ('band', )
    # Descriptor
    ProductGenerationOptions = SerializableDescriptor(
        'ProductGenerationOptions', ProductGenerationOptionsType, _required, strict=DEFAULT_STRICT,
        docstring='Performs several key actions on an image to prepare it for necessary additional processing to '
                  'achieve the desired output product.')  # type: ProductGenerationOptionsType
    RRDS = SerializableDescriptor(
        'RRDS', RRDSType, _required, strict=DEFAULT_STRICT,
        docstring='Creates a set of sub-sampled versions of an image to provide processing chains '
                  'with quick access to lower magnification values for faster computation '
                  'speeds and performance.')  # type: RRDSType
    band = IntegerDescriptor(
        'band', _required, strict=DEFAULT_STRICT,
        docstring='The image band to which this applies.')  # type: int

    def __init__(self, ProductGenerationOptions=None, RRDS=None, band=1, **kwargs):
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ProductGenerationOptions = ProductGenerationOptions
        self.RRDS = RRDS
        self.band = band
        super(NonInteractiveProcessingType, self).__init__(**kwargs)


##########
# InteractiveProcessing

class ScalingType(Serializable):
    """
    Scaling for geometric transformation
    """
    _fields = ('AntiAlias', 'Interpolation')
    _required = _fields
    # Descriptor
    AntiAlias = SerializableDescriptor(
        'AntiAlias', FilterType, _required, strict=DEFAULT_STRICT,
        docstring='The Anti-Alias Filter used for scaling. Refer to program-specific '
                  'documentation for population guidance.')  # type: FilterType
    Interpolation = SerializableDescriptor(
        'Interpolation', FilterType, _required, strict=DEFAULT_STRICT,
        docstring='The Interpolation Filter used for scaling. Refer to program-specific '
                  'documentation for population guidance.')  # type: FilterType

    def __init__(self, AntiAlias=None, Interpolation=None, **kwargs):
        """

        Parameters
        ----------
        AntiAlias : FilterType
        Interpolation : FilterType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.AntiAlias = AntiAlias
        self.Interpolation = Interpolation
        super(ScalingType, self).__init__(**kwargs)


class OrientationType(Serializable):
    """
    Parameters describing the default orientation of the product.
    """

    _fields = ('ShadowDirection', )
    _required = _fields
    # Descriptor
    ShadowDirection = StringEnumDescriptor(
        'ShadowDirection', ('UP', 'DOWN', 'LEFT', 'RIGHT', 'ARBITRARY'), _required,
        strict=DEFAULT_STRICT, default_value='DOWN',
        docstring='Describes the shadow direction relative to the '
                  'pixels in the file.')  # type: str

    def __init__(self, ShadowDirection='DOWN', **kwargs):
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ShadowDirection = ShadowDirection
        super(OrientationType, self).__init__(**kwargs)


class GeometricTransformType(Serializable):
    """
    The geometric transform element is used to perform various geometric distortions to each
    band of image data. These distortions include image chipping, scaling, rotation, shearing, etc.
    """

    _fields = ('Scaling', 'Orientation')
    _required = _fields
    # Descriptor
    Scaling = SerializableDescriptor(
        'Scaling', ScalingType, _required, strict=DEFAULT_STRICT,
        docstring='The scaling filters.')  # type: ScalingType
    Orientation = SerializableDescriptor(
        'Orientation', OrientationType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters describing the default orientation of the product.')  # type: OrientationType

    def __init__(self, Scaling=None, Orientation=None, **kwargs):
        """

        Parameters
        ----------
        Scaling : ScalingType
        Orientation : OrientationType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Scaling = Scaling
        self.Orientation = Orientation
        super(GeometricTransformType, self).__init__(**kwargs)


class SharpnessEnhancementType(Serializable):
    """
    Sharpness enhancement filter parameters.
    """
    _fields = ('ModularTransferFunctionCompensation', 'ModularTransferFunctionEnhancement')
    _required = ()
    _choice = ({'required': True, 'collection': ('ModularTransferFunctionCompensation',
                                                 'ModularTransferFunctionEnhancement')}, )
    # Descriptor
    ModularTransferFunctionCompensation = SerializableDescriptor(
        'ModularTransferFunctionCompensation', FilterType, _required, strict=DEFAULT_STRICT,
        docstring=r'If defining a custom Filter, it must be no larger than :math:`5\times 5`.')  # type: FilterType
    ModularTransferFunctionEnhancement = SerializableDescriptor(
        'ModularTransferFunctionEnhancement', FilterType, _required, strict=DEFAULT_STRICT,
        docstring=r'If defining a custom Filter, it must be no larger than :math:`5\times 5`.')  # type: FilterType

    def __init__(self, ModularTransferFunctionCompensation=None, ModularTransferFunctionEnhancement=None, **kwargs):
        """

        Parameters
        ----------
        ModularTransferFunctionCompensation : FilterType
        ModularTransferFunctionEnhancement : FilterType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ModularTransferFunctionCompensation = ModularTransferFunctionCompensation
        self.ModularTransferFunctionEnhancement = ModularTransferFunctionEnhancement
        super(SharpnessEnhancementType, self).__init__(**kwargs)


class ColorManagementModuleType(Serializable):
    """
    Parameters describing the Color Management Module (CMM).
    """

    _fields = ('RenderingIntent', 'SourceProfile', 'DisplayProfile', 'ICCProfile')
    _required = ('RenderingIntent', 'SourceProfile')
    # Descriptor
    RenderingIntent = StringEnumDescriptor(
        'RenderingIntent', ('PERCEPTUAL', 'SATURATION', 'RELATIVE', 'ABSOLUTE'),
        _required, strict=DEFAULT_STRICT, default_value='PERCEPTUAL',
        docstring='The rendering intent for this color management.')  # type: str
    SourceProfile = StringDescriptor(
        'SourceProfile', _required, strict=DEFAULT_STRICT,
        docstring='Name of sensor profile in ICC Profile database.')  # type: str
    DisplayProfile = StringDescriptor(
        'DisplayProfile', _required, strict=DEFAULT_STRICT,
        docstring='Name of display profile in ICC Profile database.')  # type: str
    ICCProfile = StringDescriptor(
        'ICCProfile', _required, strict=DEFAULT_STRICT,
        docstring='Valid ICC profile signature.')  # type: str

    def __init__(self, RenderingIntent='PERCEPTUAL', SourceProfile=None,
                 DisplayProfile=None, ICCProfile=None, **kwargs):
        """

        Parameters
        ----------
        RenderingIntent : str
        SourceProfile : str
        DisplayProfile : str
        ICCProfile : str
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RenderingIntent = RenderingIntent
        self.SourceProfile = SourceProfile
        self.DisplayProfile = DisplayProfile
        self.ICCProfile = ICCProfile
        super(ColorManagementModuleType, self).__init__(**kwargs)


class ColorSpaceTransformType(Serializable):
    """
    Parameters describing any color transformation.
    """
    _fields = ('ColorManagementModule', )
    _required = _fields
    # Descriptor
    ColorManagementModule = SerializableDescriptor(
        'ColorManagementModule', ColorManagementModuleType, _required, strict=DEFAULT_STRICT,
        docstring='Parameters describing the Color Management Module (CMM).')  # type: ColorManagementModuleType

    def __init__(self, ColorManagementModule=None, **kwargs):
        """

        Parameters
        ----------
        ColorManagementModule : ColorManagementModuleType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ColorManagementModule = ColorManagementModule
        super(ColorSpaceTransformType, self).__init__(**kwargs)


class DRAParametersType(Serializable):
    """
    The basic dynamic range adjustment parameters.
    """
    _fields = ('Pmin', 'Pmax', 'EminModifier', 'EmaxModifier')
    _required = _fields
    _numeric_format = {key: FLOAT_FORMAT for key in _fields}
    # Descriptor
    Pmin = FloatDescriptor(
        'Pmin', _required, strict=DEFAULT_STRICT, bounds=(0, 1),
        docstring='DRA clip low point. This is the cumulative histogram percentage value '
                  'that defines the lower end-point of the dynamic range to be displayed.')  # type: float
    Pmax = FloatDescriptor(
        'Pmax', _required, strict=DEFAULT_STRICT, bounds=(0, 1),
        docstring='DRA clip high point. This is the cumulative histogram percentage value '
                  'that defines the upper end-point of the dynamic range to be displayed.')  # type: float
    EminModifier = FloatDescriptor(
        'EminModifier', _required, strict=DEFAULT_STRICT, bounds=(0, 1),
        docstring='The pixel value corresponding to the Pmin percentage point in the '
                  'image histogram.')  # type: float
    EmaxModifier = FloatDescriptor(
        'EmaxModifier', _required, strict=DEFAULT_STRICT, bounds=(0, 1),
        docstring='The pixel value corresponding to the Pmax percentage point in the '
                  'image histogram.')  # type: float

    def __init__(self, Pmin=None, Pmax=None, EminModifier=None, EmaxModifier=None, **kwargs):
        """

        Parameters
        ----------
        Pmin : float
        Pmax : float
        EminModifier : float
        EmaxModifier : float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Pmin = Pmin
        self.Pmax = Pmax
        self.EminModifier = EminModifier
        self.EmaxModifier = EmaxModifier
        super(DRAParametersType, self).__init__(**kwargs)


class DRAOverridesType(Serializable):
    """
    The dynamic range adjustment overrides.
    """
    _fields = ('Subtractor', 'Multiplier')
    _required = _fields
    _numeric_format = {key: FLOAT_FORMAT for key in _fields}
    # Descriptor
    Subtractor = FloatDescriptor(
        'Subtractor', _required, strict=DEFAULT_STRICT, bounds=(0, 2047),
        docstring='Subtractor value used to reduce haze in the image.')  # type: float
    Multiplier = FloatDescriptor(
        'Multiplier', _required, strict=DEFAULT_STRICT, bounds=(0, 2047),
        docstring='Multiplier value used to reduce haze in the image.')  # type: float

    def __init__(self, Subtractor=None, Multiplier=None, **kwargs):
        """

        Parameters
        ----------
        Subtractor : float
        Multiplier : float
        kwargs
        """
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Subtractor = Subtractor
        self.Multiplier = Multiplier
        super(DRAOverridesType, self).__init__(**kwargs)


class DynamicRangeAdjustmentType(Serializable):
    """
    The dynamic range adjustment (DRA) parameters.
    """
    _fields = ('AlgorithmType', 'BandStatsSource', 'DRAParameters', 'DRAOverrides')
    _required = ('AlgorithmType', 'BandStatsSource', )
    # Descriptor
    AlgorithmType = StringEnumDescriptor(
        'AlgorithmType', ('AUTO', 'MANUAL', 'NONE'), _required, strict=DEFAULT_STRICT, default_value='NONE',
        docstring='Algorithm used for dynamic range adjustment.')  # type: str
    BandStatsSource = IntegerDescriptor(
        'BandStatsSource', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: int
    DRAParameters = SerializableDescriptor(
        'DRAParameters', DRAParametersType, _required, strict=DEFAULT_STRICT,
        docstring='The dynamic range adjustment parameters.')  # type: DRAParametersType
    DRAOverrides = SerializableDescriptor(
        'DRAOverrides', DRAOverridesType, _required, strict=DEFAULT_STRICT,
        docstring='The dynamic range adjustment overrides.')  # type: DRAOverridesType

    def __init__(self, AlgorithmType='NONE', BandStatsSource=None, DRAParameters=None, DRAOverrides=None, **kwargs):
        """

        Parameters
        ----------
        AlgorithmType : str
        BandStatsSource : int
        DRAParameters : DRAParametersType
        DRAOverrides : DRAOverridesType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.AlgorithmType = AlgorithmType
        self.BandStatsSource = BandStatsSource
        self.DRAParameters = DRAParameters
        self.DRAOverrides = DRAOverrides
        super(DynamicRangeAdjustmentType, self).__init__(**kwargs)


class InteractiveProcessingType(Serializable):
    """
    The interactive processing information.
    """
    _fields = (
        'GeometricTransform', 'SharpnessEnhancement', 'ColorSpaceTransform',
        'DynamicRangeAdjustment', 'TonalTransferCurve', 'band')
    _required = (
        'GeometricTransform', 'SharpnessEnhancement', 'DynamicRangeAdjustment', 'band')
    _set_as_attribute = ('band', )
    # Descriptor
    GeometricTransform = SerializableDescriptor(
        'GeometricTransform', GeometricTransformType, _required, strict=DEFAULT_STRICT,
        docstring='The geometric transform element is used to perform various geometric distortions '
                  'to each band of image data. These distortions include image '
                  'chipping, scaling, rotation, shearing, etc.')  # type: GeometricTransformType
    SharpnessEnhancement = SerializableDescriptor(
        'SharpnessEnhancement', SharpnessEnhancementType, _required, strict=DEFAULT_STRICT,
        docstring='Sharpness enhancement.')  # type: SharpnessEnhancementType
    ColorSpaceTransform = SerializableDescriptor(
        'ColorSpaceTransform', ColorSpaceTransformType, _required, strict=DEFAULT_STRICT,
        docstring='Color space transform.')  # type: ColorSpaceTransformType
    DynamicRangeAdjustment = SerializableDescriptor(
        'DynamicRangeAdjustment', DynamicRangeAdjustmentType, _required, strict=DEFAULT_STRICT,
        docstring='Specifies the recommended ELT DRA overrides.')  # type: DynamicRangeAdjustmentType
    TonalTransferCurve = SerializableDescriptor(
        'TonalTransferCurve', NewLookupTableType, _required, strict=DEFAULT_STRICT,
        docstring="The 1-D LUT element uses one or more 1-D LUTs to stretch or compress tone data "
                  "in valorous regions within a digital image's dynamic range. 1-D LUT can be "
                  "implemented using a Tonal Transfer Curve (TTC). There are 12 families of TTCs "
                  "- Range = [0, 11]. There are 64 members for each "
                  "family - Range=[0, 63].")  # type: NewLookupTableType
    band = IntegerDescriptor(
        'band', _required, strict=DEFAULT_STRICT,
        docstring='The image band to which this applies.')

    def __init__(self, GeometricTransform=None, SharpnessEnhancement=None,
                 ColorSpaceTransform=None, DynamicRangeAdjustment=None, band=1, **kwargs):
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.GeometricTransform = GeometricTransform
        self.SharpnessEnhancement = SharpnessEnhancement
        self.ColorSpaceTransform = ColorSpaceTransform
        self.DynamicRangeAdjustment = DynamicRangeAdjustment
        self.band = band
        super(InteractiveProcessingType, self).__init__(**kwargs)


##########

class ProductDisplayType(Serializable):
    """

    """
    _fields = (
        'PixelType', 'NumBands', 'DefaultBandDisplay', 'NonInteractiveProcessing',
        'InteractiveProcessing', 'DisplayExtensions')
    _required = (
        'PixelType', 'NumBands', 'NonInteractiveProcessing', 'InteractiveProcessing')
    _collections_tags = {
        'NonInteractiveProcessing': {'array': False, 'child_tag': 'NonInteractiveProcessing'},
        'InteractiveProcessing': {'array': False, 'child_tag': 'InteractiveProcessing'},
        'DisplayExtensions': {'array': False, 'child_tag': 'DisplayExtension'}}

    # Descriptors
    PixelType = StringEnumDescriptor(
        'PixelType', ('MONO8I', 'MONO8LU', 'MONO16I', 'RGB8LU', 'RGB24I'), _required, strict=DEFAULT_STRICT,
        docstring='Enumeration of the pixel type. Definition in '
                  'Design and Exploitation document.')  # type: str
    NumBands = IntegerDescriptor(
        'NumBands', _required, strict=DEFAULT_STRICT,
        docstring='Number of bands contained in the image. Populate with the number of bands '
                  'present after remapping. For example an 8-bit RGB image (RGBLU), this will '
                  'be 3.')  # type: int
    DefaultBandDisplay = IntegerDescriptor(
        'DefaultBandDisplay', _required, strict=DEFAULT_STRICT,
        docstring='Indicates which band to display by default. '
                  'Valid range = 1 to NumBands.')  # type: int
    NonInteractiveProcessing = SerializableListDescriptor(
        'NonInteractiveProcessing', NonInteractiveProcessingType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Non-interactive processing details.')  # type: List[NonInteractiveProcessingType]
    InteractiveProcessing = SerializableListDescriptor(
        'InteractiveProcessing', InteractiveProcessingType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Interactive processing details.')  # type: List[InteractiveProcessingType]
    DisplayExtensions = ParametersDescriptor(
        'DisplayExtensions', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Optional extensible parameters used to support profile-specific needs related to '
                  'product display. Predefined filter types.')  # type: ParametersCollection

    def __init__(self, PixelType=None, NumBands=1, DefaultBandDisplay=None,
                 NonInteractiveProcessing=None, InteractiveProcessing=None, DisplayExtensions=None, **kwargs):
        """

        Parameters
        ----------
        PixelType : PixelTypeType
        NumBands : int
        DefaultBandDisplay : int|None
        NonInteractiveProcessing : List[NonInteractiveProcessingType]
        InteractiveProcessing : List[InteractiveProcessingType]
        DisplayExtensions : ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.PixelType = PixelType
        self.NumBands = NumBands
        self.DefaultBandDisplay = DefaultBandDisplay
        self.NonInteractiveProcessing = NonInteractiveProcessing
        self.InteractiveProcessing = InteractiveProcessing
        self.DisplayExtensions = DisplayExtensions
        super(ProductDisplayType, self).__init__(**kwargs)

    def get_pixel_size(self) -> int:
        """
        Gets the raw size per pixel, in bytes.

        Returns
        -------
        int
        """

        if self.PixelType == 'MONO8I':
            return 1
        elif self.PixelType == 'MONO8LU':
            return 1
        elif self.PixelType == 'MONO16I':
            return 2
        elif self.PixelType == 'RGB8LU':
            return 1
        elif self.PixelType == 'RGB24I':
            return 3
        else:
            raise ValueError('Got unhandled pixel type `{}`'.format(self.PixelType))



"""
Provides a class for calculating the height of the WGS84 geoid above the ellipsoid
at any given latitude and longitude.

**Some Accuracy Details:**
Using the 5 minute pgm and linear interpolation, the average error `|calculated - real|`
is around 5 millimeters, and the worst case error is around 30 centimeters. Using cubic
interpolation, the average error drops to about 3 mm, and the worst case is about 17 cm.

Using the 1 minute pgm and linear interpolation, the average error is around 0.5 mm and
worst case error around 1 cm. Using cubic interpolation, the average error is still around
0.5 mm, and worst case error around 2 mm.

The accuracy obtained using the 5, 2.5, or 1 minute pgm are likely all more than suitable
for any SAR application. The the accuracy clearly increases with finer grid.

**Some Processing Speed and Resource Details:**
A memory map into the pgm file is established, which requires relatively little "real" RAM,
but an amount of virtual memory on par with the file size. The speed seems to generally scale
close to inverse linearly with pixel size, so using the 5 minute pgm is generally ~4-5 times
faster than using the 1 minute pgm.

Using the 5 minute pgm, processing rate of around 4-8 million points per second for linear
interpolation and 1-2 million points per second for cubic interpolation. Using the 1 minute pgm,
this processing rate drops to 1-2 million points per second for linear interpolation, and
~million points per second using the cubic interpolation. These rates depend of a variety
of factors including processor speed, hard drive speed, and how your operating system handles
memory maps.

The 5 minute pgm is about 25 times smaller at around 18 MB, while the 1 minute pgm file is
around 450 MB.

**File Locations:**
As of January 2020, the egm2008 pgm files are available for download at
https://geographiclib.sourceforge.io/html/geoid.html

Specifically 1 minute data is available at
https://sourceforge.net/projects/geographiclib/files/geoids-distrib/egm2008-1.tar.bz2
or
https://sourceforge.net/projects/geographiclib/files/geoids-distrib/egm2008-1.zip

Specifically the 5 minute data is available at
https://sourceforge.net/projects/geographiclib/files/geoids-distrib/egm2008-5.tar.bz2
or
https://sourceforge.net/projects/geographiclib/files/geoids-distrib/egm2008-5.zip
"""

import os
import numpy

from sarpy.io.DEM.utils import argument_validation
from sarpy.io.general.base import SarpyIOError


__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


############
# module variables
_C0 = 240
_C3 = numpy.array((
    (9,   -18, -88,    0,  96,   90,   0,   0, -60, -20),
    (-9,   18,   8,    0, -96,   30,   0,   0,  60, -20),
    (9,   -88, -18,   90,  96,    0, -20, -60,   0,   0),
    (186, -42, -42, -150, -96, -150,  60,  60,  60,  60),
    (54,  162, -78,   30, -24,  -90, -60,  60, -60,  60),
    (-9,  -32,  18,   30,  24,    0,  20, -60,   0,   0),
    (-9,    8,  18,   30, -96,    0, -20,  60,   0,   0),
    (54,  -78, 162,  -90, -24,   30,  60, -60,  60, -60),
    (-54,  78,  78,   90, 144,   90, -60, -60, -60, -60),
    (9,    -8, -18,  -30, -24,    0,  20,  60,   0,   0),
    (-9,   18, -32,    0,  24,   30,   0,   0, -60,  20),
    (9,   -18,  -8,    0, -24,  -30,   0,   0,  60,  20)), dtype=numpy.float64)

_C0N = 372
_C3N = numpy.array((
    (0, 0, -131, 0, 138, 144, 0, 0, -102, -31),
    (0, 0, 7, 0, -138, 42, 0, 0, 102, -31),
    (62, 0, -31, 0, 0, -62, 0, 0, 0, 31),
    (124, 0, -62, 0, 0, -124, 0, 0, 0, 62),
    (124, 0, -62, 0, 0, -124, 0, 0, 0, 62),
    (62, 0, -31, 0, 0, -62, 0, 0, 0, 31),
    (0, 0, 45, 0, -183, -9, 0, 93, 18, 0),
    (0, 0, 216, 0, 33, 87, 0, -93, 12, -93),
    (0, 0, 156, 0, 153, 99, 0, -93, -12, -93),
    (0, 0, -45, 0, -3, 9, 0, 93, -18, 0),
    (0, 0, -55, 0, 48, 42, 0, 0, -84, 31),
    (0, 0, -7, 0, -48, -42, 0, 0, 84, 31)), dtype=numpy.float64)

_C0S = 372
_C3S = numpy.array((
    (18, -36, -122, 0, 120, 135, 0, 0, -84, -31),
    (-18, 36, -2, 0, -120, 51, 0, 0, 84, -31),
    (36, -165, -27, 93, 147, -9, 0, -93, 18, 0),
    (210, 45, -111, -93, -57, -192, 0, 93, 12, 93),
    (162, 141, -75, -93, -129, -180, 0, 93, -12, 93),
    (-36, -21, 27, 93, 39, 9, 0, -93, -18, 0),
    (0, 0, 62, 0, 0, 31, 0, 0, 0, -31),
    (0, 0, 124, 0, 0, 62, 0, 0, 0, -62),
    (0, 0, 124, 0, 0, 62, 0, 0, 0, -62),
    (0, 0, 62, 0, 0, 31, 0, 0, 0, -31),
    (-18, 36, -64, 0, 66, 51, 0, 0, -102, 31),
    (18, -36, 2, 0, -66, -51, 0, 0, 102, 31)), dtype=numpy.float64)

_SEARCH_FILES = ('egm2008-5.pgm', 'egm2008-2_5.pgm', 'egm2008-1.pgm', 'egm96-5.pgm', 'egm96-15.pgm')


def find_geoid_file_from_dir(dir_name, search_files=None):
    """
    Find the geoid file.

    Parameters
    ----------
    dir_name : str
    search_files : str|List[str]

    Returns
    -------
    str
    """

    geoid_dir = os.path.join(dir_name, 'geoid')
    if not os.path.exists(geoid_dir):
        raise SarpyIOError(
            'Input is a directory, and beneath it we expect to find '
            'files in directory "geoid"')
    if search_files is None:
        search_files = []
    elif isinstance(search_files, str):
        search_files = [search_files, ]
    else:
        search_files = list(search_files)

    for entry in _SEARCH_FILES:
        if entry not in search_files:
            search_files.append(entry)

    our_file = None
    for fil in search_files:
        file_name = os.path.join(geoid_dir, fil)
        if os.path.exists(file_name):
            our_file = file_name
            break

    if our_file is None:
        raise SarpyIOError(
            'input is a directory and we expect to find one of the files {} '
            'in the directory "geoid" beneath it'.format(search_files))

    return our_file


class GeoidHeight(object):
    """
    Calculator for the height of the WGS84 geoid above the ellipsoid at any
    given latitude and longitude, based on one of the egm .pgm files.

    We are set up to use a dem/geoid parent directory. In this case, we expect
    our egm .pgm to be in the `<root_dir>/geoid` directory, and we will search
    in order of preference
    ('egm2008-1.pgm', 'egm2008-2_5.pgm', 'egm2008-5.pgm', 'egm96-5.pgm', 'egm96-15.pgm')
    """

    __slots__ = (
        '_offset', '_scale', '_width', '_height', '_header_length', '_memory_map',
        '_lon_res', '_lat_res')

    def __init__(self, file_name):
        """

        Parameters
        ----------
        file_name : str
            path to a egm2008 pgm file
        """

        self._offset = None
        self._scale = None

        if os.path.isdir(file_name):
            file_name = find_geoid_file_from_dir(file_name)

        with open(file_name, "rb") as f:
            line = f.readline()
            if line != b"P5\012" and line != b"P5\015\012":
                raise SarpyIOError("No PGM header")
            headerlen = len(line)
            while True:
                line = f.readline().decode('utf-8')
                if len(line) == 0:
                    raise SarpyIOError("EOF before end of file header")
                headerlen += len(line)
                if line.startswith('# Offset '):
                    try:
                        self._offset = int(line[9:])
                    except ValueError as e:
                        raise SarpyIOError("Error reading offset", e)
                elif line.startswith('# Scale '):
                    try:
                        self._scale = float(line[8:])
                    except ValueError as e:
                        raise SarpyIOError("Error reading scale", e)
                elif not line.startswith('#'):
                    try:
                        slin = line.split()
                        self._width, self._height = int(slin[0]), int(slin[1])
                    except ValueError as e:
                        raise SarpyIOError("Bad PGM width&height line", e)
                    break
            line = f.readline().decode('utf-8')
            headerlen += len(line)
            levels = int(line)
            if levels != 65535:
                raise SarpyIOError("PGM file must have 65535 gray levels")
            if self._offset is None:
                raise SarpyIOError("PGM file does not contain offset")
            if self._scale is None:
                raise SarpyIOError("PGM file does not contain scale")

            if self._width < 2 or self._height < 2:
                raise SarpyIOError("Raster size too small")
            self._header_length = headerlen

        self._memory_map = numpy.memmap(file_name,
                                        dtype=numpy.dtype('>u2'),
                                        mode='r',
                                        offset=self._header_length,
                                        shape=(self._height, self._width))
        self._lon_res = self._width/360.0
        self._lat_res = (self._height - 1)/180.0

    def _get_raw(self, ix, iy):
        # these manipulations are required for edge effects
        boolc = (iy < 0)
        iy[boolc] *= -1
        # ix[boolc] += int(self._width/2)  # why is this here?

        boolc = (iy >= self._height)
        iy[boolc] = 2*(self._height - 1) - iy[boolc]
        # ix[boolc] += int(self._width/2)  # why is this here?

        boolc = (ix < 0)
        ix[boolc] += self._width

        boolc = (ix >= self._width)
        ix[boolc] -= self._width

        return self._memory_map[iy, ix]

    def _linear(self, ix, dx, iy, dy):
        a = (1 - dx) * self._get_raw(ix, iy) + dx * self._get_raw(ix + 1, iy)
        b = (1 - dx) * self._get_raw(ix, iy+1) + dx * self._get_raw(ix+1, iy+1)
        return (1 - dy) * a + dy * b

    def _cubic(self, ix, dx, iy, dy):
        v = numpy.vstack((
            self._get_raw(ix, iy - 1),
            self._get_raw(ix + 1, iy - 1),
            self._get_raw(ix - 1, iy),
            self._get_raw(ix, iy),
            self._get_raw(ix + 1, iy),
            self._get_raw(ix + 2, iy),
            self._get_raw(ix - 1, iy + 1),
            self._get_raw(ix, iy + 1),
            self._get_raw(ix + 1, iy + 1),
            self._get_raw(ix + 2, iy + 1),
            self._get_raw(ix, iy + 2),
            self._get_raw(ix + 1, iy + 2)))

        t = numpy.zeros((10, ix.size), dtype=numpy.float64)
        b1 = (iy == 0)
        b2 = (iy == self._height - 2)
        b3 = ~(b1 | b2)
        if numpy.any(b1):
            t[:, b1] = (_C3N.T/_C0N).dot(v[:, b1])
        if numpy.any(b2):
            t[:, b2] = (_C3S.T/_C0S).dot(v[:, b2])
        if numpy.any(b3):
            t[:, b3] = (_C3.T/_C0).dot(v[:, b3])

        return t[0] + \
            dx*(t[1] + dx*(t[3] + dx*t[6])) + \
            dy*(t[2] + dx*(t[4] + dx*t[7]) + dy*(t[5] + dx*t[8] + dy*t[9]))

    def _do_block(self, lat, lon, cubic):
        fx = lon*self._lon_res
        fx[fx < 0] += 360*self._lon_res
        fy = (90 - lat)*self._lat_res

        ix = numpy.asarray(numpy.floor(fx), dtype=numpy.int32)
        iy = numpy.asarray(numpy.floor(fy), dtype=numpy.int32)

        dx = fx - ix
        dy = fy - iy

        iy[iy == self._height - 1] -= 1  # edge effects?

        if cubic:
            return self._offset + self._scale*self._cubic(ix, dx, iy, dy)
        else:
            return self._offset + self._scale*self._linear(ix, dx, iy, dy)

    def get(self, lat, lon, cubic=True, block_size=50000):
        """
        Calculate the height of the geoid above the ellipsoid in meters at the given points.

        Parameters
        ----------
        lat : numpy.ndarray
        lon : numpy.ndarray
        cubic : bool
            Use a simple cubic spline interpolation, otherwise us simple linear.
            Default is `True`.
        block_size : None|int
            If `None`, then the entire calculation will proceed as a single block.
            Otherwise, block processing using blocks of the given size will be used.
            The minimum value used for this is 50,000, and any smaller value will be
            replaced with 50,000. Default is 50,000.

        Returns
        -------
        numpy.ndarray
        """

        o_shape, lat, lon = argument_validation(lat, lon)

        if block_size is None:
            out = self._do_block(lat, lon, cubic)
        else:
            block_size = max(50000, int(block_size))
            out = numpy.empty(lat.shape, dtype=numpy.float64)
            start_block = 0
            while start_block < lat.size:
                end_block = min(start_block+block_size, lat.size)
                out[start_block:end_block] = self._do_block(
                    lat[start_block:end_block], lon[start_block:end_block], cubic)
                start_block = end_block

        if o_shape == ():
            return float(out[0])
        else:
            return numpy.reshape(out, o_shape)

    def __call__(self, lat, lon):
        return self.get(lat, lon)

    @classmethod
    def from_directory(cls, dir_name, search_files=None):
        """
        Create the GeoidHeight object from a search directory.

        Parameters
        ----------
        dir_name : str
        search_files : str|List[str]

        Returns
        -------
        GeoidHeight
        """

        our_file = find_geoid_file_from_dir(dir_name, search_files=search_files)
        return cls(our_file)



"""
The SCPCOAType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
from typing import Union, Optional, Tuple

import numpy
from numpy.linalg import norm

from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import StringEnumDescriptor, FloatDescriptor, \
    SerializableDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import XYZType

from sarpy.geometry import geocoords

logger = logging.getLogger(__name__)


class GeometryCalculator(object):
    """
    Performs the necessary SCPCOA geometry element calculations.
    """

    def __init__(
            self,
            SCP: numpy.ndarray,
            ARPPos: numpy.ndarray,
            ARPVel: numpy.ndarray):
        """

        Parameters
        ----------
        SCP : numpy.ndarray
            The scene center point.
        ARPPos : numpy.ndarray
            The aperture position in ECEF coordinates at the SCP center of aperture time.
        ARPVel : numpy.ndarray
            The aperture velocity in ECEF coordinates at the SCP center of aperture time.
        """

        self.SCP = SCP
        self.ARP = ARPPos
        self.ARP_vel = ARPVel
        self.LOS = (self.SCP - self.ARP)
        # unit vector versions
        self.uSCP = self._make_unit(self.SCP)
        self.uARP = self._make_unit(self.ARP)
        self.uARP_vel = self._make_unit(self.ARP_vel)
        self.uLOS = self._make_unit(self.LOS)
        self.left = numpy.cross(self.uARP, self.uARP_vel)
        self.look = numpy.sign(self.left.dot(self.uLOS))
        # Earth Tangent Plane (ETP) at the SCP is the plane tangent to the surface of constant height
        # above the WGS 84 ellipsoid (HAE) that contains the SCP. The ETP is an approximation to the
        # ground plane at the SCP.
        self.ETP = geocoords.wgs_84_norm(SCP)
        # slant plane unit normal
        self.uSPZ = self._make_unit(self.look*numpy.cross(self.ARP_vel, self.uLOS))
        # perpendicular component of range vector wrt the ground plane
        self.uGPX = self._make_unit(-self.uLOS + numpy.dot(self.ETP, self.uLOS)*self.ETP)
        self.uGPY = numpy.cross(self.ETP, self.uGPX)  # already unit vector
        # perpendicular component of north wrt the ground plane
        self.uNORTH = self._make_unit(numpy.array([0, 0, 1]) - self.ETP[2]*self.ETP)
        self.uEAST = numpy.cross(self.uNORTH, self.ETP)  # already unit vector

    @staticmethod
    def _make_unit(vec: numpy.ndarray) -> numpy.ndarray:
        vec_norm = norm(vec)
        if vec_norm < 1e-6:
            logger.error(
                'The input vector to be normalized has norm {},\n\t'
                'this is likely a mistake'.format(vec_norm))
        return vec/vec_norm

    @property
    def ROV(self) -> float:
        """
        float: Range over velocity
        """
        return float(norm(self.LOS)/norm(self.ARP_vel))

    @property
    def SideOfTrack(self) -> str:
        return 'R' if self.look < 0 else 'L'

    @property
    def SlantRange(self) -> float:
        return float(norm(self.LOS))

    @property
    def GroundRange(self) -> float:
        return norm(self.SCP)*numpy.arccos(self.uSCP.dot(self.uARP))

    @property
    def DopplerConeAng(self) -> float:
        return float(numpy.rad2deg(numpy.arccos(self.uARP_vel.dot(self.uLOS))))

    @property
    def GrazeAng(self) -> float:
        return self.get_graze_and_incidence()[0]

    @property
    def IncidenceAng(self) -> float:
        return self.get_graze_and_incidence()[1]

    def get_graze_and_incidence(self) -> Tuple[float, float]:
        graze_ang = -float(numpy.rad2deg(numpy.arcsin(self.ETP.dot(self.uLOS))))
        return graze_ang, 90 - graze_ang

    @property
    def TwistAng(self) -> float:
        return float(-numpy.rad2deg(numpy.arcsin(self.uGPY.dot(self.uSPZ))))

    @property
    def SquintAngle(self) -> float:
        # Note: squint angle is not defined in the SICD DIDD, but it is defined in the SIDD DIDD v2 (7.5.4)
        #     squint_angle = arccos(-Xs_hat dot Va_hat), where "hat" means unit vector.
        # In this code, uLOS = -xs_hat, so the minus sign is omitted from the argument of the arccos function.
        # It is also stated in table 4.2.5 that "Left-look is positive, right-look is negative".
        arp_vel_proj = self._make_unit(self.uARP_vel - self.uARP_vel.dot(self.uARP)*self.uARP)
        los_proj = self._make_unit(self.uLOS - self.uLOS.dot(self.uARP)*self.uARP)
        return float(self.look * numpy.rad2deg(numpy.arccos(los_proj.dot(arp_vel_proj))))

    @property
    def SlopeAng(self) -> float:
        return float(numpy.rad2deg(numpy.arccos(self.ETP.dot(self.uSPZ))))

    @property
    def AzimAng(self) -> float:
        azim_ang = numpy.rad2deg(numpy.arctan2(self.uGPX.dot(self.uEAST), self.uGPX.dot(self.uNORTH)))
        azim_ang = azim_ang if azim_ang > 0 else azim_ang + 360
        return float(azim_ang)

    @property
    def LayoverAng(self) -> float:
        return self.get_layover()[0]

    def get_layover(self) -> Tuple[float, float]:
        layover_ground = self.ETP - self.ETP.dot(self.uSPZ)*self.uSPZ
        layover_ang = numpy.rad2deg(
            numpy.arctan2(layover_ground.dot(self.uEAST), layover_ground.dot(self.uNORTH)))
        layover_ang = layover_ang if layover_ang > 0 else layover_ang + 360
        return float(layover_ang), float(norm(layover_ground))

    def get_shadow(self) -> Tuple[float, float]:
        shadow = self.ETP - self.uLOS/self.uLOS.dot(self.ETP)
        shadow_prime = shadow - self.uSPZ*(shadow.dot(self.ETP)/self.uSPZ.dot(self.ETP))
        shadow_angle = numpy.rad2deg(numpy.arctan2(shadow_prime.dot(self.uGPY), shadow_prime.dot(self.uGPX)))
        return float(shadow_angle), float(norm(shadow_prime))


class SCPCOAType(Serializable):
    """
    Center of Aperture (COA) for the Scene Center Point (SCP).
    """

    _fields = (
        'SCPTime', 'ARPPos', 'ARPVel', 'ARPAcc', 'SideOfTrack', 'SlantRange', 'GroundRange', 'DopplerConeAng',
        'GrazeAng', 'IncidenceAng', 'TwistAng', 'SlopeAng', 'AzimAng', 'LayoverAng')
    _required = _fields
    _numeric_format = {
        'SCPTime': FLOAT_FORMAT, 'SlantRange': '0.17E', 'GroundRange': '0.17E',
        'DopplerConeAng': FLOAT_FORMAT, 'GrazeAng': FLOAT_FORMAT, 'IncidenceAng': FLOAT_FORMAT,
        'TwistAng': FLOAT_FORMAT, 'SlopeAng': FLOAT_FORMAT, 'AzimAng': FLOAT_FORMAT,
        'LayoverAng': FLOAT_FORMAT}
    # class variables
    _SIDE_OF_TRACK_VALUES = ('L', 'R')
    # descriptors
    SCPTime = FloatDescriptor(
        'SCPTime', _required, strict=DEFAULT_STRICT,
        docstring='*Center Of Aperture time for the SCP (t_COA_SCP)*, relative to collection '
                  'start in seconds.')  # type: float
    ARPPos = SerializableDescriptor(
        'ARPPos', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='Aperture position at *t_COA_SCP* in ECF coordinates.')  # type: XYZType
    ARPVel = SerializableDescriptor(
        'ARPVel', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='ARP Velocity at *t_COA_SCP* in ECF coordinates.')  # type: XYZType
    ARPAcc = SerializableDescriptor(
        'ARPAcc', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='ARP Acceleration at *t_COA_SCP* in ECF coordinates.')  # type: XYZType
    SideOfTrack = StringEnumDescriptor(
        'SideOfTrack', _SIDE_OF_TRACK_VALUES, _required, strict=DEFAULT_STRICT,
        docstring='Side of track.')  # type: str
    SlantRange = FloatDescriptor(
        'SlantRange', _required, strict=DEFAULT_STRICT,
        docstring='Slant range from the aperture to the *SCP* in meters.')  # type: float
    GroundRange = FloatDescriptor(
        'GroundRange', _required, strict=DEFAULT_STRICT,
        docstring='Ground Range from the aperture nadir to the *SCP*. Distance measured along spherical earth model '
                  'passing through the *SCP* in meters.')  # type: float
    DopplerConeAng = FloatDescriptor(
        'DopplerConeAng', _required, strict=DEFAULT_STRICT,
        docstring='The Doppler Cone Angle to SCP at *t_COA_SCP* in degrees.')  # type: float
    GrazeAng = FloatDescriptor(
        'GrazeAng', _required, strict=DEFAULT_STRICT, bounds=(0., 90.),
        docstring='Grazing Angle between the SCP *Line of Sight (LOS)* and *Earth Tangent Plane (ETP)*.')  # type: float
    IncidenceAng = FloatDescriptor(
        'IncidenceAng', _required, strict=DEFAULT_STRICT, bounds=(0., 90.),
        docstring='Incidence Angle between the *LOS* and *ETP* normal.')  # type: float
    TwistAng = FloatDescriptor(
        'TwistAng', _required, strict=DEFAULT_STRICT, bounds=(-90., 90.),
        docstring='Angle between cross range in the *ETP* and cross range in the slant plane.')  # type: float
    SlopeAng = FloatDescriptor(
        'SlopeAng', _required, strict=DEFAULT_STRICT, bounds=(0., 90.),
        docstring='Slope Angle from the *ETP* to the slant plane at *t_COA_SCP*.')  # type: float
    AzimAng = FloatDescriptor(
        'AzimAng', _required, strict=DEFAULT_STRICT, bounds=(0., 360.),
        docstring='Angle from north to the line from the *SCP* to the aperture nadir at *COA*. Measured '
                  'clockwise in the *ETP*.')  # type: float
    LayoverAng = FloatDescriptor(
        'LayoverAng', _required, strict=DEFAULT_STRICT, bounds=(0., 360.),
        docstring='Angle from north to the layover direction in the *ETP* at *COA*. Measured '
                  'clockwise in the *ETP*.')  # type: float

    def __init__(
            self,
            SCPTime: float = None,
            ARPPos: Union[XYZType, numpy.ndarray, list, tuple] = None,
            ARPVel: Union[XYZType, numpy.ndarray, list, tuple] = None,
            ARPAcc: Union[XYZType, numpy.ndarray, list, tuple] = None,
            SideOfTrack: str = None,
            SlantRange: float = None,
            GroundRange: float = None,
            DopplerConeAng: float = None,
            GrazeAng: float = None,
            IncidenceAng: float = None,
            TwistAng: float = None,
            SlopeAng: float = None,
            AzimAng: float = None,
            LayoverAng: float = None,
            **kwargs):
        """

        Parameters
        ----------
        SCPTime : float
        ARPPos : XYZType|numpy.ndarray|list|tuple
        ARPVel : XYZType|numpy.ndarray|list|tuple
        ARPAcc : XYZType|numpy.ndarray|list|tuple
        SideOfTrack : str
        SlantRange : float
        GroundRange : float
        DopplerConeAng : float
        GrazeAng : float
        IncidenceAng : float
        TwistAng : float
        SlopeAng : float
        AzimAng : float
        LayoverAng : float
        kwargs
        """

        self._ROV = None
        self._squint = None
        self._shadow = None
        self._shadow_magnitude = None
        self._layover_magnitude = None

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.SCPTime = SCPTime
        self.ARPPos, self.ARPVel, self.ARPAcc = ARPPos, ARPVel, ARPAcc
        self.SideOfTrack = SideOfTrack
        self.SlantRange, self.GroundRange = SlantRange, GroundRange
        self.DopplerConeAng, self.GrazeAng, self.IncidenceAng = DopplerConeAng, GrazeAng, IncidenceAng
        self.TwistAng, self.SlopeAng, self.AzimAng, self.LayoverAng = TwistAng, SlopeAng, AzimAng, LayoverAng
        super(SCPCOAType, self).__init__(**kwargs)

    @property
    def look(self) -> Optional[int]:
        """
        int: An integer version of `SideOfTrack`:

            * None if `SideOfTrack` is not defined

            * -1 if SideOfTrack == 'R'

            * 1 if SideOftrack == 'L'
        """

        if self.SideOfTrack is None:
            return None
        else:
            return -1 if self.SideOfTrack == 'R' else 1

    @property
    def ROV(self) -> Optional[float]:
        """
        float: The Ratio of Range to Velocity at Center of Aperture time.
        """

        return self._ROV

    @property
    def ThetaDot(self) -> Optional[float]:
        """
        float: Derivative of Theta as a function of time at Center of Aperture time.
        """

        if self.DopplerConeAng is None or self.ROV is None:
            return None
        return float(numpy.sin(numpy.deg2rad(self.DopplerConeAng))/self.ROV)

    @property
    def MultipathGround(self) -> Optional[float]:
        """
        float: The anticipated angle of multipath features on the ground in degrees.
        """
        if self.GrazeAng is None or self.TwistAng is None:
            return None
        return numpy.rad2deg(
            -numpy.arctan(numpy.tan(numpy.deg2rad(self.TwistAng))*numpy.sin(numpy.deg2rad(self.GrazeAng))))

    @property
    def Multipath(self) -> Optional[float]:
        """
        float: The anticipated angle of multipath features in degrees.
        """
        if self.AzimAng is None or self.MultipathGround is None:
            return None
        return numpy.mod(self.AzimAng - 180 + self.MultipathGround, 360)

    @property
    def Shadow(self) -> Optional[float]:
        """
        float: The anticipated angle of shadow features in degrees.
        """

        return self._shadow

    @property
    def ShadowMagnitude(self) -> Optional[float]:
        """
        float: The anticipated relative magnitude of shadow features.
        """

        return self._shadow_magnitude

    @property
    def Squint(self) -> float:
        """
        float: The squint angle, in degrees.
        """

        return self._squint

    @property
    def LayoverMagnitude(self) -> float:
        """
        float: The anticipated relative magnitude of layover features.
        """

        return self._layover_magnitude

    def _derive_scp_time(self, Grid, overwrite: bool = False):
        """
        Expected to be called by SICD parent.

        Parameters
        ----------
        Grid : sarpy.io.complex.sicd_elements.GridType
        overwrite : bool

        Returns
        -------
        None
        """

        if Grid is None or Grid.TimeCOAPoly is None:
            return  # nothing can be done
        if not overwrite and self.SCPTime is not None:
            return  # nothing should be done

        scp_time = Grid.TimeCOAPoly.Coefs[0, 0]
        self.SCPTime = scp_time

    def _derive_position(self, Position, overwrite: bool = False):
        """
        Derive aperture position parameters, if necessary. Expected to be called by SICD parent.

        Parameters
        ----------
        Position : sarpy.io.complex.sicd_elements.Position.PositionType
        overwrite : bool

        Returns
        -------
        None
        """

        if Position is None or Position.ARPPoly is None or self.SCPTime is None:
            return  # nothing can be derived

        # set aperture position, velocity, and acceleration at scptime from position
        # polynomial, if necessary
        poly = Position.ARPPoly
        scptime = self.SCPTime

        if self.ARPPos is None or overwrite:
            self.ARPPos = XYZType.from_array(poly(scptime))
            self.ARPVel = XYZType.from_array(poly.derivative_eval(scptime, 1))
            self.ARPAcc = XYZType.from_array(poly.derivative_eval(scptime, 2))

    def _derive_geometry_parameters(self, GeoData, overwrite: bool = False):
        """
        Expected to be called by SICD parent.

        Parameters
        ----------
        GeoData : sarpy.io.complex.sicd_elements.GeoData.GeoDataType
        overwrite : bool

        Returns
        -------
        None
        """

        if GeoData is None or GeoData.SCP is None or GeoData.SCP.ECF is None or \
                self.ARPPos is None or self.ARPVel is None:
            return  # nothing can be derived

        # construct our calculator
        calculator = GeometryCalculator(
            GeoData.SCP.ECF.get_array(), self.ARPPos.get_array(), self.ARPVel.get_array())
        # set all the values
        self._ROV = calculator.ROV
        if self.SideOfTrack is None or overwrite:
            self.SideOfTrack = calculator.SideOfTrack
        if self.SlantRange is None or overwrite:
            self.SlantRange = calculator.SlantRange
        if self.GroundRange is None or overwrite:
            self.GroundRange = calculator.GroundRange
        if self.DopplerConeAng is None or overwrite:
            self.DopplerConeAng = calculator.DopplerConeAng
        graz, inc = calculator.get_graze_and_incidence()
        if self.GrazeAng is None or overwrite:
            self.GrazeAng = graz
        if self.IncidenceAng is None or overwrite:
            self.IncidenceAng = inc
        if self.TwistAng is None or overwrite:
            self.TwistAng = calculator.TwistAng
        self._squint = calculator.SquintAngle
        if self.SlopeAng is None or overwrite:
            self.SlopeAng = calculator.SlopeAng
        if self.AzimAng is None or overwrite:
            self.AzimAng = calculator.AzimAng
        layover, self._layover_magnitude = calculator.get_layover()
        if self.LayoverAng is None or overwrite:
            self.LayoverAng = layover
        self._shadow, self._shadow_magnitude = calculator.get_shadow()

    def rederive(self, Grid, Position, GeoData):
        """
        Rederive all derived quantities.

        Parameters
        ----------
        Grid : sarpy.io.complex.sicd_elements.GridType
        Position : sarpy.io.complex.sicd_elements.Position.PositionType
        GeoData : sarpy.io.complex.sicd_elements.GeoData.GeoDataType

        Returns
        -------
        None
        """

        self._derive_scp_time(Grid, overwrite=True)
        self._derive_position(Position, overwrite=True)
        self._derive_geometry_parameters(GeoData, overwrite=True)

    def check_values(self, GeoData) -> bool:
        """
        Check derived values for validity.

        Parameters
        ----------
        GeoData : sarpy.io.complex.sicd_elements.GeoData.GeoDataType

        Returns
        -------
        bool
        """

        if GeoData is None or GeoData.SCP is None or GeoData.SCP.ECF is None or \
                self.ARPPos is None or self.ARPVel is None:
            return True  # nothing can be derived

        # construct our calculator
        calculator = GeometryCalculator(
            GeoData.SCP.ECF.get_array(), self.ARPPos.get_array(), self.ARPVel.get_array())

        cond = True
        if calculator.SideOfTrack != self.SideOfTrack:
            self.log_validity_error(
                'SideOfTrack is expected to be {}, and is populated as {}'.format(
                    calculator.SideOfTrack, self.SideOfTrack))
            cond = False

        for attribute in ['SlantRange', 'GroundRange']:
            val1 = getattr(self, attribute)
            val2 = getattr(calculator, attribute)
            if abs(val1/val2 - 1) > 1e-6:
                self.log_validity_error(
                    'attribute {} is expected to have value {}, but is populated as {}'.format(attribute, val2, val1))
                cond = False

        for attribute in [
                'DopplerConeAng', 'GrazeAng', 'IncidenceAng', 'TwistAng', 'SlopeAng', 'AzimAng', 'LayoverAng']:
            val1 = getattr(self, attribute)
            val2 = getattr(calculator, attribute)
            if abs(val1 - val2) > 1e-3:
                self.log_validity_error(
                    'attribute {} is expected to have value {}, but is populated as {}'.format(attribute, val2, val1))
                cond = False
        return cond




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class CSDIDAType(TREElement):
    def __init__(self, value):
        super(CSDIDAType, self).__init__()
        self.add_field('DAY', 's', 2, value)
        self.add_field('MONTH', 's', 3, value)
        self.add_field('YEAR', 's', 4, value)
        self.add_field('PLATFORM_CODE', 's', 2, value)
        self.add_field('VEHICLE_ID', 's', 2, value)
        self.add_field('PASS', 's', 2, value)
        self.add_field('OPERATION', 's', 3, value)
        self.add_field('SENSOR_ID', 's', 2, value)
        self.add_field('PRODUCT_ID', 's', 2, value)
        self.add_field('RESERVED_1', 's', 4, value)
        self.add_field('TIME', 's', 14, value)
        self.add_field('PROCESS_TIME', 's', 14, value)
        self.add_field('RESERVED_2', 's', 2, value)
        self.add_field('RESERVED_3', 's', 2, value)
        self.add_field('RESERVED_4', 's', 1, value)
        self.add_field('RESERVED_5', 's', 1, value)
        self.add_field('SOFTWARE_VERSION_NUMBER', 's', 10, value)


class CSDIDA(TREExtension):
    _tag_value = 'CSDIDA'
    _data_type = CSDIDAType



"""
The text extension subheader definitions.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from .base import NITFElement, UserHeaderType, _IntegerDescriptor,\
    _StringDescriptor, _StringEnumDescriptor, _NITFElementDescriptor
from .security import NITFSecurityTags, NITFSecurityTags0


########
# NITF 2.1

class TextSegmentHeader(NITFElement):
    """
    Text Segment Subheader for NITF version 2.1 - see standards document
    Joint BIIF Profile (JBP) for more information.
    """

    _ordering = (
        'TE', 'TEXTID', 'TXTALVL', 'TXTDT', 'TXTITL', 'Security',
        'ENCRYP', 'TXTFMT', 'UserHeader')
    _lengths = {
        'TE': 2, 'TEXTID': 7, 'TXTALVL': 3, 'TXTDT': 14, 'TXTITL': 80,
        'ENCRYP': 1, 'TXTFMT': 3}
    TE = _StringEnumDescriptor(
        'TE', True, 2, {'TE', }, default_value='TE',
        docstring='File part type.')  # type: str
    TEXTID = _StringDescriptor(
        'TEXTID', True, 7, default_value='',
        docstring='Text Identifier. This field shall contain a valid alphanumeric identification '
                  'code associated with the text item. The valid codes are determined '
                  'by the application.')  # type: str
    TXTALVL = _IntegerDescriptor(
        'TXTALVL', True, 3, default_value=0,
        docstring='Text Attachment Level. This field shall contain a valid value that '
                  'indicates the attachment level of the text.')  # type: int
    TXTDT = _StringDescriptor(
        'TXTDT', True, 14, default_value='',
        docstring='Text Date and Time. This field shall contain the time (UTC) of origination '
                  'of the text in the format :code:`YYYYMMDDhhmmss`')  # type: str
    TXTITL = _StringDescriptor(
        'TXTITL', True, 80, default_value='',
        docstring='Text Title. This field shall contain the title of the text item.')  # type: str
    Security = _NITFElementDescriptor(
        'Security', True, NITFSecurityTags, default_args={},
        docstring='The security tags.')  # type: NITFSecurityTags
    ENCRYP = _StringEnumDescriptor(
        'ENCRYP', True, 1, {'0'}, default_value='0',
        docstring='Encryption.')  # type: str
    TXTFMT = _StringEnumDescriptor(
        'TXTFMT', True, 3, {'', 'MTF', 'STA', 'UT1', 'U8S'}, default_value='',
        docstring='Text Format. This field shall contain a valid three character code '
                  'indicating the format or type of text data. Valid codes are :code:`MTF` to '
                  'indicate USMTF (Refer to MIL-STD-6040 for examples of the USMTF format), '
                  ':code:`STA` to indicate BCS, :code:`UT1` to indicate ECS text formatting, and '
                  ':code:`U8S` to indicate U8S text formatting.')  # type: str
    UserHeader = _NITFElementDescriptor(
        'UserHeader', True, UserHeaderType, default_args={},
        docstring='User defined header.')  # type: UserHeaderType


########
# NITF 2.0

class TextSegmentHeader0(NITFElement):
    """
    Text Segment Subheader for NITF version 2.0 - see standards document
    MIL-STD-2500A for more information.
    """

    _ordering = (
        'TE', 'TEXTID', 'TXTDT', 'TXTITL', 'Security',
        'ENCRYP', 'TXTFMT', 'UserHeader')
    _lengths = {
        'TE': 2, 'TEXTID': 10, 'TXTDT': 14, 'TXTITL': 80,
        'ENCRYP': 1, 'TXTFMT': 3}
    TE = _StringEnumDescriptor(
        'TE', True, 2, {'TE', }, default_value='TE',
        docstring='File part type.')  # type: str
    TEXTID = _StringDescriptor(
        'TEXTID', True, 10, default_value='',
        docstring='Text Identifier. This field shall contain a valid alphanumeric identification '
                  'code associated with the text item. The valid codes are determined '
                  'by the application.')  # type: str
    TXTDT = _StringDescriptor(
        'TXTDT', True, 14, default_value='',
        docstring='Text Date and Time. This field shall contain the time (UTC) of origination '
                  'of the text in the format :code:`YYYYMMDDhhmmss`')  # type: str
    TXTITL = _StringDescriptor(
        'TXTITL', True, 80, default_value='',
        docstring='Text Title. This field shall contain the title of the text item.')  # type: str
    Security = _NITFElementDescriptor(
        'Security', True, NITFSecurityTags0, default_args={},
        docstring='The security tags.')  # type: NITFSecurityTags0
    ENCRYP = _StringEnumDescriptor(
        'ENCRYP', True, 1, {'0'}, default_value='0',
        docstring='Encryption.')  # type: str
    TXTFMT = _StringDescriptor(
        'TXTFMT', True, 3, default_value='',
        docstring='Text Format. This field shall contain a valid three character code '
                  'indicating the format or type of text data. Valid codes are :code:`MTF` to '
                  'indicate USMTF (Refer to MIL-STD-6040 for examples of the USMTF format), '
                  ':code:`STA` to indicate BCS, :code:`UT1` to indicate ECS text formatting, and '
                  ':code:`U8S` to indicate U8S text formatting.')  # type: str
    UserHeader = _NITFElementDescriptor(
        'UserHeader', True, UserHeaderType, default_args={},
        docstring='User defined header.')  # type: UserHeaderType



"""
Extract information from the CPHD header for review.

From the command-line

>>> python -m sarpy.utils.cphd_utils <path to cphd file>

For a basic help on the command-line, check

>>> python -m sarpy.utils.cphd_utils --help

"""

from __future__ import print_function
import argparse
import sys
import functools
from xml.dom import minidom
from typing import Union, TextIO, BinaryIO
import os
from io import StringIO

from sarpy.io.phase_history.cphd import CPHDDetails

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

# Custom print function
print_func = print


def _define_print_function(destination):
    """
    Define the print_func as necessary.

    Parameters
    ----------
    destination : TextIO
    """

    global print_func
    print_func = functools.partial(print, file=destination)


def _print_header(input_file):
    # type: (Union[str, BinaryIO]) -> None

    def _finalize():
        if close_after:
            file_object.close()
        if initial_location is not None:
            file_object.seek(initial_location)

    if isinstance(input_file, str):
        file_object = open(input_file, 'rb')
        initial_location = None
        close_after = True
    elif hasattr(input_file, 'readline'):
        file_object = input_file
        initial_location = file_object.tell()
        file_object.seek(0)
        close_after = False
    else:
        raise TypeError(
            'Input requires a file path or a binary mode file-like object')

    while True:
        lin = file_object.readline().strip()
        if not isinstance(lin, bytes):
            _finalize()
            raise ValueError('Requires an input opened in binary mode')

        if lin:
            print_func(lin.decode())
        else:
            break
    _finalize()


def _create_default_output_file(input_file):
    # type: (Union[str, BinaryIO]) -> str
    if isinstance(input_file, str):
        return os.path.splitext(input_file)[0] + '.meta_dump.txt'
    else:
        return os.path.expanduser('~/Desktop/phase_history.meta_dump.txt')


def _print_structure(input_file):
    # type: (Union[str, BinaryIO]) -> None
    details = CPHDDetails(input_file)
    data = details.get_cphd_bytes()
    xml_str = minidom.parseString(data.decode()).toprettyxml(indent='    ', newl='\n')
    # NB: this may or not exhibit platform dependent choices in which codec (i.e. latin-1 versus utf-8)
    for i, entry in enumerate(xml_str.splitlines()):
        if i == 0:
            # Remove xml that gets inserted by minidom, if it's not actually there
            if (not data.startswith(b'<?xml version')) and entry.startswith('<?xml version'):
                continue
            print_func(entry)
        elif entry.strip() != '':
            # Remove extra new lines if XML is already formatted
            print_func(entry)


def print_cphd_metadata(input_file, destination=sys.stdout):
    """
    Prints the full CPHD metadata (both header and CPHD structure) to the
    given destination.

    Parameters
    ----------
    input_file : str|BinaryIO
    destination : TextIO
    """

    _define_print_function(destination)

    if isinstance(input_file, str):
        print_func('Details for CPHD file {}'.format(input_file))

    print_func('---- CPHD Header Information ----')
    _print_header(input_file)
    print_func('')
    print_func('')

    print_func('---- CPHD Structure ----')
    _print_structure(input_file)
    print_func('')


def print_cphd_header(input_file, destination=sys.stdout):
    """
    Prints the full CPHD header to the given destination.

    Parameters
    ----------
    input_file : str|BinaryIO
    destination : TextIO
    """

    _define_print_function(destination)
    _print_header(input_file)


def print_cphd_xml(input_file, destination=sys.stdout):
    """
    Prints the full CPHD header to the given destination.

    Parameters
    ----------
    input_file : str|BinaryIO
    destination : TextIO
    """

    _define_print_function(destination)
    _print_structure(input_file)


def _dump_pattern(input_file, destination, call_method):
    # type: (Union[str, BinaryIO], str, Callable) -> Union[None, str]
    if destination == 'stdout':
        call_method(input_file, destination=sys.stdout)
    elif destination == 'string':
        out = StringIO()
        call_method(input_file, destination=out)
        value = out.getvalue()
        out.close()  # free the buffer
        return value
    else:
        the_out_file = _create_default_output_file(input_file) if destination == 'default' else destination
        with open(the_out_file, 'w') as fi:
            call_method(input_file, destination=fi)


def dump_cphd_metadata(input_file, destination):
    """
    Dump the CPHD metadata (both header and CPHD structure) to the given
    destination.

    Parameters
    ----------
    input_file : str|BinaryIO
        Path to or binary file-like object containing a CPHD file.
    destination : str
        'stdout', 'string', 'default' (will use `file_name+'.meta_dump.txt'`),
        or the path to an output file.

    Returns
    -------
    None|str
        There is only a return value if `destination=='string'`.
    """

    _dump_pattern(input_file, destination, print_cphd_metadata)


def dump_cphd_header(input_file, destination):
    """
    Dump the CPHD header to the given destination.

    Parameters
    ----------
    input_file : str|BinaryIO
        Path to or binary file-like object containing a CPHD file.
    destination : str
        'stdout', 'string', 'default' (will use `file_name+'.meta_dump.txt'`),
        or the path to an output file.

    Returns
    -------
    None|str
        There is only a return value if `destination=='string'`.
    """

    _dump_pattern(input_file, destination, print_cphd_header)


def dump_cphd_xml(input_file, destination):
    """
    Dump the CPHD structure to the given destination.

    Parameters
    ----------
    input_file : str|BinaryIO
        Path to or binary file-like object containing a CPHD file.
    destination : str
        'stdout', 'string', 'default' (will use `file_name+'.meta_dump.txt'`),
        or the path to an output file.

    Returns
    -------
    None|str
        There is only a return value if `destination=='string'`.
    """

    _dump_pattern(input_file, destination, print_cphd_xml)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description="Create extract metadata information from a CPHD file.",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument(
        'input_file', metavar='input_file', help='Path input CPHD file.')
    parser.add_argument(
        '-o', '--output', default='default',
        help="'default', 'stdout', or the path for an output file.\n"
             "* If not provided (`default`), the output will be at '<input path>.txt' \n"
             "* 'stdout' will print the information to standard out.\n"
             "* Otherwise, this is expected to be the path to a file \n"
             "       and output will be written there.\n"
             "  NOTE: existing output files will be overwritten.")
    parser.add_argument(
        '-d', '--data', default='both', choices=['both', 'header', 'xml'],
        help='Which information should be printed?')
    args = parser.parse_args()

    if args.data == 'both':
        dump_cphd_metadata(args.input_file, args.output)
    elif args.data == 'header':
        dump_cphd_header(args.input_file, args.output)
    elif args.data == 'xml':
        dump_cphd_xml(args.input_file, args.output)
    else:
        raise ValueError('Got unhandled data option {}'.format(args.data))




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class PIAIMBType(TREElement):
    def __init__(self, value):
        super(PIAIMBType, self).__init__()
        self.add_field('CLOUD', 's', 3, value)
        self.add_field('STDRD', 's', 1, value)
        self.add_field('SMODE', 's', 12, value)
        self.add_field('SNAME', 's', 18, value)
        self.add_field('SRCE', 's', 255, value)
        self.add_field('CMGEN', 's', 2, value)
        self.add_field('SQUAL', 's', 1, value)
        self.add_field('MISNM', 's', 7, value)
        self.add_field('CSPEC', 's', 32, value)
        self.add_field('PJTID', 's', 2, value)
        self.add_field('GENER', 's', 1, value)
        self.add_field('EXPLS', 's', 1, value)
        self.add_field('OTHRC', 's', 2, value)


class PIAIMB(TREExtension):
    _tag_value = 'PIAIMB'
    _data_type = PIAIMBType




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class PATCHBType(TREElement):
    def __init__(self, value):
        super(PATCHBType, self).__init__()
        self.add_field('PAT_NO', 's', 4, value)
        self.add_field('LAST_PAT_FLAG', 's', 1, value)
        self.add_field('LNSTRT', 's', 7, value)
        self.add_field('LNSTOP', 's', 7, value)
        self.add_field('AZL', 's', 5, value)
        self.add_field('NVL', 's', 5, value)
        self.add_field('FVL', 's', 3, value)
        self.add_field('NPIXEL', 's', 5, value)
        self.add_field('FVPIX', 's', 5, value)
        self.add_field('FRAME', 's', 3, value)
        self.add_field('UTC', 's', 8, value)
        self.add_field('SHEAD', 's', 7, value)
        self.add_field('GRAVITY', 's', 7, value)
        self.add_field('INS_V_NC', 's', 5, value)
        self.add_field('INS_V_EC', 's', 5, value)
        self.add_field('INS_V_DC', 's', 5, value)
        self.add_field('OFFLAT', 's', 8, value)
        self.add_field('OFFLONG', 's', 8, value)
        self.add_field('TRACK', 's', 3, value)
        self.add_field('GSWEEP', 's', 6, value)
        self.add_field('SHEAR', 's', 8, value)
        self.add_field('BATCH_NO', 's', 6, value)


class PATCHB(TREExtension):
    _tag_value = 'PATCHB'
    _data_type = PATCHBType



"""
Module contained elements for defining TREs - really intended as read only objects.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import functools
import logging
import struct
from collections import OrderedDict
from typing import Union, List

from ..base import TRE

logger = logging.getLogger(__name__)


def _parse_type(typ_string, leng, value, start):
    """

    Parameters
    ----------
    typ_string : str
    leng : int
    value : bytes
    start : int

    Returns
    -------
    str|int|bytes|float
    """

    byt = value[start:start + leng]
    if typ_string == 's':
        return byt.decode('utf-8').strip()
    if typ_string == 'd':
        return int(byt)
    if typ_string == 'ieee754_binary32':
        return struct.unpack('>f', byt)[0]
    if typ_string == 'b':
        return byt
    raise ValueError(f'Got unrecognized type string {typ_string}')


def _str_encoder(val, formatspec):
    return f'{val:{formatspec}}'.encode('utf-8')


def _create_encoder(typ_string, leng):
    if typ_string == 's':
        return functools.partial(_str_encoder, formatspec=f'{leng}s')
    if typ_string == 'd':
        return functools.partial(_str_encoder, formatspec=f'0{leng}d')
    if typ_string == 'b':
        return lambda x: x
    if typ_string == 'ieee754_binary32':
        return functools.partial(struct.pack, '>f')


class TREElement(object):
    """
    Basic TRE element class
    """

    def __init__(self):
        self._field_ordering = []
        self._field_encoders = {}
        self._bytes_length = 0

    def __str__(self):
        return '{0:s}({1:s})'.format(self.__class__.__name__, self.to_dict())

    def __repr__(self):
        return '{0:s}(b"'.format(self.__class__.__name__) + self.to_bytes().decode() + '")'

    def add_field(self, attribute, typ_string, leng, value):
        """
        Add a field/attribute to the object - as we deserialize.

        Parameters
        ----------
        attribute : str
            The new field/attribute name for out object instance.
        typ_string : str
            One of 's' (string), 'd' (integer), 'b' (raw/bytes), or 'ieee754_binary32'
        leng : int
            The length in bytes of the representation of this attribute
        value : bytes
            The bytes array of the object we are deserializing

        Returns
        -------
        None
        """

        if hasattr(self, attribute):
            logger.error(
                'This instance of TRE element {} already has an attribute {},\n\t'
                'but the `add_field()` method is being called for this attribute name again.\n\t'
                'This is almost certainly an error.'.format(self.__class__, attribute))

        try:
            val = _parse_type(typ_string, leng, value, self._bytes_length)
            setattr(self, attribute, val)
        except Exception as e:
            raise ValueError(
                'Failed creating field {} with exception \n\t{}'.format(attribute, e))
        self._bytes_length += leng
        self._field_ordering.append(attribute)
        self._field_encoders[attribute] = _create_encoder(typ_string, leng)

    def add_loop(self, attribute, length, child_type, value, *args):
        """
        Add an attribute from a loop construct of a given type to the object - as we deserialize.

        Parameters
        ----------
        attribute : str
            The new field/attribute name for out object instance.
        length : int
            The number of loop iterations present.
        child_type : type
            The type of the child - must extend TREElement
        value : bytes
            The bytes array of the object we are deserializing
        args
            Any optional positional arguments that the child_type constructor should have.

        Returns
        -------
        None
        """

        try:
            obj = TRELoop(length, child_type, value, self._bytes_length, *args)
            setattr(self, attribute, obj)
        except Exception as e:
            raise ValueError(
                'Failed creating loop {} of type {} with exception\n\t{}'.format(attribute, child_type, e))
        self._bytes_length += obj.get_bytes_length()
        self._field_ordering.append(attribute)

    def _attribute_to_bytes(self, attribute):
        """
        Get byte representation for the given attribute.

        Parameters
        ----------
        attribute : str

        Returns
        -------
        bytes
        """

        val = getattr(self, attribute, None)
        if val is None:
            return b''
        elif isinstance(val, TREElement):
            return val.to_bytes()
        else:
            return self._field_encoders[attribute](val)

    def to_dict(self):
        """
        Create a dictionary representation of the object.

        Returns
        -------
        dict
        """

        out = OrderedDict()
        for fld in self._field_ordering:
            val = getattr(self, fld)
            if val is None or isinstance(val, (bytes, str, int, float)):
                out[fld] = val
            elif isinstance(val, TREElement):
                out[fld] = val.to_dict()
            else:
                raise TypeError('Unhandled type {}'.format(type(val)))
        return out

    def get_bytes_length(self):
        """
        The length in bytes of the serialized representation.

        Returns
        -------
        int
        """

        return self._bytes_length

    def to_bytes(self):
        """
        Serialize to bytes.

        Returns
        -------
        bytes
        """

        items = [self._attribute_to_bytes(fld) for fld in self._field_ordering]
        return b''.join(items)

    def to_json(self):
        """
        Gets a json representation of this element.

        Returns
        -------
        dict|list
        """

        out = OrderedDict()
        for fld in self._field_ordering:
            value = getattr(self, fld)
            if isinstance(value, TREElement):
                out[fld] = value.to_json()
            else:
                out[fld] = value


class TRELoop(TREElement):
    """
    Provides the TRE loop construct
    """

    def __init__(self, length, child_type, value, start, *args, **kwargs):
        """

        Parameters
        ----------
        length : int
        child_type : type
        value : bytes
        start : int
        args
            optional positional args for child class construction
        kwargs
            optional keyword arguments for child class construction
        """

        if not issubclass(child_type, TREElement):
            raise TypeError('child_class must be a subclass of TREElement.')

        super(TRELoop, self).__init__()
        self._data = []
        loc = start
        for i in range(length):
            entry = child_type(value[loc:], *args, **kwargs)
            leng = entry.get_bytes_length()
            self._bytes_length += leng
            loc += leng
            self._data.append(entry)

    def to_dict(self):
        return [entry.to_dict() for entry in self._data]

    def to_bytes(self):
        return b''.join(entry.to_bytes() for entry in self._data)

    def __len__(self):
        return len(self._data)

    def __getitem__(self, item):  # type: (Union[int, slice]) -> Union[TREElement, List[TREElement]]
        return self._data[item]

    def to_json(self):
        """
        Gets a json representation of this element.

        Returns
        -------
        dict|list
        """

        return [entry.to_json() for entry in self._data]


class TREExtension(TRE):
    """
    Extend this object to provide concrete TRE implementations.
    """

    __slots__ = ('_data', )
    _tag_value = None
    _data_type = None

    def __init__(self, value):
        if not issubclass(self._data_type, TREElement):
            raise TypeError('_data_type must be a subclass of TREElement. Got type {}'.format(self._data_type))
        if not isinstance(self._tag_value, str):
            raise TypeError('_tag_value must be a string')
        if len(self._tag_value) > 6:
            raise ValueError('Tag value must have 6 or fewer characters.')
        self._data = None
        self.DATA = value

    @property
    def TAG(self):
        return self._tag_value

    @property
    def DATA(self):  # type: () -> _data_type
        return self._data

    @DATA.setter
    def DATA(self, value):
        # type: (Union[bytes, _data_type]) -> None
        if isinstance(value, self._data_type):
            self._data = value
        elif isinstance(value, bytes):
            self._data = self._data_type(value)
        else:
            raise TypeError(
                'data must be of {} type or a bytes array. '
                'Got {}'.format(self._data_type, type(value)))

    @property
    def EL(self):
        if self._data is None:
            return 0
        return self._data.get_bytes_length()

    @classmethod
    def minimum_length(cls):
        return 11

    def get_bytes_length(self):
        return 11 + self.EL

    def to_bytes(self):
        return ('{0:6s}{1:05d}'.format(self.TAG, self.EL)).encode('utf-8') + self._data.to_bytes()

    @classmethod
    def from_bytes(cls, value, start):
        tag_value = value[start:start+6].decode('utf-8').strip()
        lng = int(value[start+6:start+11])
        if tag_value != cls._tag_value:
            raise ValueError('tag value must be {}. Got {}'.format(cls._tag_value, tag_value))
        return cls(value[start+11:start+11+lng])



__classification__ = 'UNCLASSIFIED'
__author__ = "Thomas McCullough"

import os
import re
import logging
from typing import List, Dict, Optional


logger = logging.getLogger('validation')

_the_directory = os.path.split(__file__)[0]
urn_mapping = {
    'urn:SIDD:1.0.0': {
        'ism_urn': 'urn:us:gov:ic:ism',
        'sfa_urn': 'urn:SFA:1.2.0',
        'sicommon_urn': 'urn:SICommon:0.1',
        'version': '1.0.0',
        'release': '1.0.0',
        'date': '2011-08-31T00:00:00Z',
        'schema': os.path.join(_the_directory, 'version1', 'SIDD_schema_V1.0.0_2011_08_31.xsd')},
    'urn:SIDD:2.0.0': {
        'ism_urn': 'urn:us:gov:ic:ism:13',
        'sfa_urn': 'urn:SFA:1.2.0',
        'sicommon_urn': 'urn:SICommon:1.0',
        'version': '2.0.0',
        'release': '2.0.0',
        'date': '2019-05-31T00:00:00Z',
        'schema': os.path.join(_the_directory, 'version2', 'SIDD_schema_V2.0.0_2019_05_31.xsd')},
    'urn:SIDD:3.0.0': {
        'ism_urn': 'urn:us:gov:ic:ism:13',
        'sfa_urn': 'urn:SFA:1.2.0',
        'sicommon_urn': 'urn:SICommon:1.0',
        'version': '3.0.0',
        'release': '3.0.0',
        'date': '2020-06-02T00:00:00Z',
        'schema': os.path.join(_the_directory, 'version3', 'SIDD_schema_V3.0.0.xsd')},
}
_SIDD_SPECIFICATION_IDENTIFIER = 'SIDD Volume 1 Design & Implementation Description Document'


def get_specification_identifier() -> str:
    """
    Get the SIDD specification identifier string.

    Returns
    -------
    str
    """

    return _SIDD_SPECIFICATION_IDENTIFIER


def check_urn(urn_string: str) -> str:
    """
    Checks that the urn string follows the correct pattern.

    Parameters
    ----------
    urn_string : str

    Returns
    -------
    str

    Raises
    ------
    ValueError
        This raises an exception for a poorly formed or unmapped SIDD urn.
    """

    if not isinstance(urn_string, str):
        raise TypeError(
            'Expected a urn input of string type, got type {}'.format(type(urn_string)))

    the_match = re.match(r'^\d.\d.\d$', urn_string)
    if the_match is not None:
        urn_string = 'urn:SIDD:{}'.format(urn_string)

    the_match = re.match(r'^urn:SIDD:\d.\d.\d$', urn_string)
    if the_match is None:
        raise ValueError(
            'Input provided as `{}`,\nbut should be of the form '
            '`urn:SIDD:<major>.<minor>.<release>'.format(urn_string))
    return urn_string


def get_urn_details(urn_string: str) -> Dict[str, str]:
    """
    Gets the associated details for the given SIDD urn, or raise an exception for
    poorly formatted or unrecognized urn.

    Parameters
    ----------
    urn_string : str

    Returns
    -------
    Dict[str, str]
    """

    urn_string = check_urn(urn_string)
    out = urn_mapping.get(urn_string, None)

    if out is None:
        raise KeyError(
            'Got correctly formatted, but unmapped SIDD urn {}.'.format(urn_string))
    return out


def get_schema_path(the_urn: str) -> Optional[str]:
    """
    Gets the path to the proper schema file for the given SIDD urn.

    Parameters
    ----------
    the_urn : str

    Returns
    -------
    None|str
    """

    result = get_urn_details(the_urn)
    return result.get('schema', None)


def get_versions() -> List[str]:
    """
    Gets a list of recognized SIDD urn.

    Returns
    -------
    List[str]
    """

    return list(sorted(urn_mapping.keys()))


def validate_xml_ns(xml_ns: Dict[str, str], ns_key: str = 'default') -> bool:
    """
    Validate the parsed SIDD xml namespace dictionary. This is expected to
    accompany the use of :func:`sarpy.io.general.utils.parse_xml_from_string`.

    Parameters
    ----------
    xml_ns : dict
        The xml namespace dictionary.
    ns_key : str
        The main SIDD element or default namespace.

    Returns
    -------
    bool
    """

    if not isinstance(xml_ns, dict):
        raise ValueError('xml_ns must be a dictionary for SIDD interpretation.')

    if ns_key not in xml_ns:
        raise ValueError('ns_key must be a key in xml_ns.')

    sidd_urn = xml_ns[ns_key]

    try:
        details = get_urn_details(sidd_urn)
    except KeyError:
        logger.error('Got unmapped sidd urn `{}`'.format(sidd_urn))
        return False

    # key => (expected prefix, required)
    expected_ns = {
        'ism': ('urn:us:gov:ic:ism', True),
        'sfa': ('urn:sfa:', False),
        'sicommon': ('urn:sicommon:', True),
    }

    ns_to_add = dict()
    for expected_key, (expected_prefix, _) in expected_ns.items():
        if expected_key not in xml_ns:
            for actual_ns in xml_ns.values():
                if isinstance(actual_ns, str) and actual_ns.lower().startswith(expected_prefix):
                    ns_to_add[expected_key] = actual_ns
                    break
    xml_ns.update(ns_to_add)

    valid = True
    for key, (_, required) in expected_ns.items():
        if key in xml_ns and xml_ns[key] != details[f'{key}_urn']:
            valid = False
            logger.error(
                'SIDD: SIDD {} `{}` namespace urn is expected to be "{}", but we got "{}".\n\t'
                'Differences in standard may lead to deserialization and/or '
                'validation errors.'.format(sidd_urn, key, details[f'{key}_urn'], xml_ns[key]))
        if required and key not in xml_ns:
            valid = False
            logger.error(f'SIDD: No `{key}` namespace defined.')
    return valid



"""
Module for reading and writing SICD files
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import re
import logging
import datetime
from typing import BinaryIO, Union, Optional, Dict, Tuple, Sequence

import numpy

from sarpy.__about__ import __title__, __version__
from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd_elements.ImageCreation import ImageCreationType

from sarpy.io.general.base import SarpyIOError
from sarpy.io.general.format_function import FormatFunction, ComplexFormatFunction
from sarpy.io.general.nitf import NITFDetails, NITFReader, NITFWriter, \
    interpolate_corner_points_string, default_image_segmentation, \
    ImageSubheaderManager, TextSubheaderManager, DESSubheaderManager, \
    RESSubheaderManager, NITFWritingDetails
from sarpy.io.general.nitf_elements.nitf_head import NITFHeader
from sarpy.io.general.nitf_elements.des import DataExtensionHeader, XMLDESSubheader
from sarpy.io.general.nitf_elements.security import NITFSecurityTags
from sarpy.io.general.nitf_elements.image import ImageSegmentHeader, \
    ImageSegmentHeader0, ImageBands, ImageBand
from sarpy.io.general.utils import is_file_like

from sarpy.io.xml.base import parse_xml_from_string

logger = logging.getLogger(__name__)


#########
# Helper object for initially parses NITF header


class AmpLookupFunction(ComplexFormatFunction):
    __slots__ = ('_magnitude_lookup_table', )
    _allowed_ordering = ('MP', )

    def __init__(
            self,
            raw_dtype: Union[str, numpy.dtype],
            magnitude_lookup_table: numpy.ndarray,
            raw_shape: Optional[Tuple[int, ...]] = None,
            formatted_shape: Optional[Tuple[int, ...]] = None,
            reverse_axes: Optional[Tuple[int, ...]] = None,
            transpose_axes: Optional[Tuple[int, ...]] = None,
            band_dimension: int = -1):
        """

        Parameters
        ----------
        raw_dtype : str|numpy.dtype
            The raw datatype. Must be `uint8` up to endianness.
        magnitude_lookup_table : numpy.ndarray
        raw_shape : None|Tuple[int, ...]
        formatted_shape : None|Tuple[int, ...]
        reverse_axes : None|Tuple[int, ...]
        transpose_axes : None|Tuple[int, ...]
        band_dimension : int
            Which band is the complex dimension, **after** the transpose operation.
        """

        ComplexFormatFunction.__init__(
            self, raw_dtype, 'MP', raw_shape=raw_shape, formatted_shape=formatted_shape,
            reverse_axes=reverse_axes, transpose_axes=transpose_axes, band_dimension=band_dimension)
        self._magnitude_lookup_table = None
        self.set_magnitude_lookup(magnitude_lookup_table)

    @property
    def magnitude_lookup_table(self) -> numpy.ndarray:
        """
        The magnitude lookup table, for SICD usage with `AMP8I_PHS8I` pixel type.

        Returns
        -------
        numpy.ndarray
        """

        return self._magnitude_lookup_table

    def set_magnitude_lookup(self, lookup_table: numpy.ndarray) -> None:
        if not isinstance(lookup_table, numpy.ndarray):
            raise ValueError('requires a numpy.ndarray, got {}'.format(type(lookup_table)))
        if lookup_table.dtype.name not in ['float32', 'float64']:
            raise ValueError('requires a numpy.ndarray of float32 or 64 dtype, got {}'.format(lookup_table.dtype))
        if lookup_table.dtype.name != 'float32':
            lookup_table = numpy.asarray(lookup_table, dtype=numpy.float32)
        if lookup_table.shape != (256,):
            raise ValueError('Requires a one-dimensional numpy.ndarray with 256 elements, '
                             'got shape {}'.format(lookup_table.shape))

        if self._raw_dtype.name != 'uint8':
            raise ValueError(
                'A magnitude lookup table has been supplied,\n\t'
                'but the raw datatype is not `uint8`.')
        self._magnitude_lookup_table = lookup_table

    def _forward_magnitude_theta(
            self,
            data: numpy.ndarray,
            out: numpy.ndarray,
            magnitude: numpy.ndarray,
            theta: numpy.ndarray,
            subscript: Tuple[slice, ...]) -> None:
        magnitude = self.magnitude_lookup_table[magnitude]
        ComplexFormatFunction._forward_magnitude_theta(
            self, data, out, magnitude, theta, subscript)

    def _reverse_magnitude_theta(
            self,
            data: numpy.ndarray,
            out: numpy.ndarray,
            magnitude: numpy.ndarray,
            theta: numpy.ndarray,
            slice0: Tuple[slice, ...],
            slice1: Tuple[slice, ...]) -> None:
        magnitude = numpy.digitize(
            numpy.round(magnitude.ravel()), self.magnitude_lookup_table, right=False).reshape(data.shape)

        ComplexFormatFunction._reverse_magnitude_theta(self, data, out, magnitude, theta, slice0, slice1)


class SICDDetails(NITFDetails):
    """
    SICD are stored in NITF 2.1 files.
    """
    __slots__ = (
        '_des_index', '_des_header', '_is_sicd', '_sicd_meta')

    def __init__(self, file_object: Union[str, BinaryIO]):
        """

        Parameters
        ----------
        file_object : str|BinaryIO
            file name or file like object for a NITF 2.1 or 2.0 containing a SICD.
        """

        self._des_index = None
        self._des_header = None
        self._img_headers = None
        self._is_sicd = False
        self._sicd_meta = None
        NITFDetails.__init__(self, file_object)

        if self._nitf_header.ImageSegments.subhead_sizes.size == 0:
            raise SarpyIOError('There are no image segments defined.')
        if self._nitf_header.GraphicsSegments.item_sizes.size > 0:
            raise SarpyIOError('A SICD file does not allow for graphics segments.')
        if self._nitf_header.DataExtensions.subhead_sizes.size == 0:
            raise SarpyIOError(
                'A SICD file requires at least one data extension, containing the '
                'SICD xml structure.')

        # define the sicd metadata
        self._find_sicd()
        if not self.is_sicd:
            raise SarpyIOError('Could not find the SICD XML des.')

    @property
    def is_sicd(self) -> bool:
        """
        bool: whether file name corresponds to a SICD file, or not.
        """

        return self._is_sicd

    @property
    def sicd_meta(self) -> SICDType:
        """
        SICDType: the sicd meta-data structure.
        """

        return self._sicd_meta

    @property
    def des_header(self) -> Optional[DataExtensionHeader]:
        """
        The DES subheader object associated with the SICD.

        Returns
        -------
        DataExtensionHeader
        """

        return self._des_header

    def _find_sicd(self) -> None:
        self._is_sicd = False
        self._sicd_meta = None
        if self.des_subheader_offsets is None:
            return

        for i in range(self.des_subheader_offsets.size):
            subhead_bytes = self.get_des_subheader_bytes(i)
            if subhead_bytes.startswith(b'DEXML_DATA_CONTENT'):
                des_header = DataExtensionHeader.from_bytes(subhead_bytes, start=0)
                des_bytes = self.get_des_bytes(i)
                # noinspection PyBroadException
                try:
                    root_node, xml_ns = parse_xml_from_string(des_bytes.decode('utf-8').strip().encode())
                    if 'SIDD' in root_node.tag:  # namespace makes this ugly
                        # NOTE that SIDD files are supposed to have the corresponding
                        # SICD xml as one of the DES AFTER the SIDD xml.
                        # The same basic format is used for both headers.
                        # So, abandon if we find a SIDD xml
                        self._des_index = None
                        self._des_header = None
                        self._is_sicd = False
                        break
                    elif 'SICD' in root_node.tag:  # namespace makes this ugly
                        self._des_index = i
                        self._des_header = des_header
                        self._is_sicd = True
                        if xml_ns is None:
                            self._sicd_meta = SICDType.from_node(root_node, xml_ns, ns_key=None)
                        else:
                            self._sicd_meta = SICDType.from_node(root_node, xml_ns, ns_key='default')
                        break
                except Exception:
                    continue
            elif subhead_bytes.startswith(b'DESIDD_XML'):
                # This is an old format SIDD and can't be a SICD
                self._des_index = None
                self._des_header = None
                self._is_sicd = False
                break
            elif subhead_bytes.startswith(b'DESICD_XML'):
                # This is an old format SICD
                des_bytes = self.get_des_bytes(i)
                try:
                    root_node, xml_ns = parse_xml_from_string(des_bytes.decode('utf-8').strip().encode())
                    if 'SICD' in root_node.tag:  # namespace makes this ugly
                        self._des_index = i
                        self._des_header = None
                        self._is_sicd = True
                        if xml_ns is None:
                            self._sicd_meta = SICDType.from_node(root_node, xml_ns, ns_key=None)
                        else:
                            self._sicd_meta = SICDType.from_node(root_node, xml_ns, ns_key='default')
                        break
                except Exception as e:
                    logger.error(
                        'We found an apparent old-style SICD DES header,\n\t'
                        'but failed parsing with error {}'.format(e))
                    continue

        if not self._is_sicd:
            return

        # noinspection PyBroadException
        try:
            self._sicd_meta.derive()
        except Exception:
            pass
        # TODO: account for the reference frequency offset situation


#######
#  The actual reading implementation

class SICDReader(NITFReader, SICDTypeReader):
    """
    A SICD reader implementation - file is NITF container following specified rules.

    **Changed in version 1.3.0** for reading changes.
    """
    _maximum_number_of_images = 1

    def __init__(self, nitf_details):
        """

        Parameters
        ----------
        nitf_details :  : str|BinaryIO|SICDDetails
            filename, file-like object, or SICDDetails object
        """

        if isinstance(nitf_details, str) or is_file_like(nitf_details):
            nitf_details = SICDDetails(nitf_details)
        if not isinstance(nitf_details, SICDDetails):
            raise TypeError(
                'The input argument for SICDReader must be a filename, file-like object, '
                'or SICDDetails object.')

        SICDTypeReader.__init__(self, None, nitf_details.sicd_meta)
        NITFReader.__init__(self, nitf_details, reader_type='SICD')
        self._check_sizes()

    @property
    def nitf_details(self) -> SICDDetails:
        """
        SICDDetails: The SICD NITF details object.
        """

        # noinspection PyTypeChecker
        return self._nitf_details

    def get_nitf_dict(self) -> Dict:
        """
        Populate a dictionary with the pertinent NITF header information. This
        is for use in more faithful preservation of NITF header information
        in copying or rewriting sicd files.

        Returns
        -------
        dict
        """

        out = {}
        security = {}
        security_obj = self.nitf_details.nitf_header.Security
        # noinspection PyProtectedMember
        for field in NITFSecurityTags._ordering:
            value = getattr(security_obj, field).strip()
            if value != '':
                security[field] = value
        if len(security) > 0:
            out['Security'] = security

        out['OSTAID'] = self.nitf_details.nitf_header.OSTAID
        out['FTITLE'] = self.nitf_details.nitf_header.FTITLE
        out['ISORCE'] = self.nitf_details.img_headers[0].ISORCE
        out['IID2'] = self.nitf_details.img_headers[0].IID2
        return out

    def populate_nitf_information_into_sicd(self):
        """
        Populate some pertinent NITF header information into the SICD structure.
        This provides more faithful copying or rewriting options.
        """

        self._sicd_meta.NITF = self.get_nitf_dict()

    def depopulate_nitf_information(self):
        """
        Eliminates the NITF information dict from the SICD structure.
        """

        self._sicd_meta.NITF = {}

    def get_format_function(
            self,
            raw_dtype: numpy.dtype,
            complex_order: Optional[str],
            lut: Optional[numpy.ndarray],
            band_dimension: int,
            image_segment_index: Optional[int] = None,
            **kwargs) -> Optional[FormatFunction]:
        if complex_order is not None and complex_order != 'IQ':
            if complex_order != 'MP' or raw_dtype.name != 'uint8' or band_dimension != 2:
                raise ValueError('Got unsupported SICD band type definition')
            if self.sicd_meta.ImageData.PixelType != 'AMP8I_PHS8I' or self.sicd_meta.ImageData.AmpTable is None:
                raise ValueError('Expected AMP8I_PHS8I')
            return AmpLookupFunction(raw_dtype, self.sicd_meta.ImageData.AmpTable)
        return NITFReader.get_format_function(
            self, raw_dtype, complex_order, lut, band_dimension, image_segment_index, **kwargs)

    def _check_image_segment_for_compliance(
            self,
            index: int,
            img_header: Union[ImageSegmentHeader, ImageSegmentHeader0]) -> bool:
        out = NITFReader._check_image_segment_for_compliance(self, index, img_header)
        if not out:
            return out

        raw_dtype, formatted_dtype, formatted_bands, complex_order, lut = self._get_dtypes(index)
        if complex_order is None or complex_order not in ['IQ', 'MP']:
            logger.error(
                'Image segment at index {} is not of appropriate type for a SICD Image Segment'.format(index))
            return False
        if formatted_bands != 1:
            logger.error(
                'Image segment at index {} has multiple complex bands'.format(index))
            return False

        raw_name = raw_dtype.name
        pixel_type = self.sicd_meta.ImageData.PixelType
        if pixel_type == 'RE32F_IM32F':
            if complex_order != 'IQ' or raw_name != 'float32':
                logger.error(
                    'Image segment at index {} required to be compatible\n\t'
                    'with PIXEL_TYPE {}'.format(index, pixel_type))
                return False
        elif pixel_type == 'RE16I_IM16I':
            if complex_order != 'IQ' or raw_name != 'int16':
                logger.error(
                    'Image segment at index {} required to be compatible\n\t'
                    'with PIXEL_TYPE {}'.format(index, pixel_type))
                return False
        elif pixel_type == 'AMP8I_PHS8I':
            if complex_order != 'MP' or raw_name != 'uint8':
                logger.error(
                    'Image segment at index {} required to be compatible\n\t'
                    'with PIXEL_TYPE {}'.format(index, pixel_type))
                return False
        else:
            raise ValueError('Unhandled PIXEL_TYPE {}'.format(pixel_type))
        return True

    def find_image_segment_collections(self) -> Tuple[Tuple[int, ...]]:
        return (
            tuple(index for index in range(len(self.nitf_details.img_headers))
                  if index not in self.unsupported_segments), )


########
# base expected functionality for a module with an implemented Reader

def is_a(file_name: Union[str, BinaryIO]) -> Optional[SICDReader]:
    """
    Tests whether a given file_name corresponds to a SICD file, and returns
    a reader instance, if so.

    Parameters
    ----------
    file_name : str
        the file_name to check

    Returns
    -------
    SICDReader|None
    """

    try:
        nitf_details = SICDDetails(file_name)
        if nitf_details.is_sicd:
            logger.info('File {} is determined to be a SICD (NITF format) file.'.format(file_name))
            return SICDReader(nitf_details)
        else:
            return None
    except SarpyIOError:
        return None


#######
#  The actual writing implementation

def validate_sicd_for_writing(sicd_meta: SICDType) -> SICDType:
    """
    Helper method which ensures the provided SICD structure provides enough
    information to support file writing, as well as ensures a few basic items
    are populated as appropriate.

    Parameters
    ----------
    sicd_meta : SICDType

    Returns
    -------
    SICDType
        This returns a deep copy of the provided SICD structure, with any
        necessary modifications.
    """

    if not isinstance(sicd_meta, SICDType):
        raise ValueError('sicd_meta is required to be an instance of SICDType, got {}'.format(type(sicd_meta)))
    if sicd_meta.ImageData is None:
        raise ValueError('The sicd_meta has un-populated ImageData, and nothing useful can be inferred.')
    if sicd_meta.ImageData.NumCols is None or sicd_meta.ImageData.NumRows is None:
        raise ValueError('The sicd_meta has ImageData with unpopulated NumRows or NumCols, '
                         'and nothing useful can be inferred.')
    if sicd_meta.ImageData.PixelType is None:
        logger.warning('The PixelType for sicd_meta is unset, so defaulting to RE32F_IM32F.')
        sicd_meta.ImageData.PixelType = 'RE32F_IM32F'

    sicd_meta = sicd_meta.copy()

    profile = '{} {}'.format(__title__, __version__)
    # use naive datetime because numpy warns about parsing timezone aware
    now = numpy.datetime64(datetime.datetime.now(tz=datetime.timezone.utc).replace(tzinfo=None))
    if sicd_meta.ImageCreation is None:
        sicd_meta.ImageCreation = ImageCreationType(
            Application=profile,
            DateTime=now,
            Profile=profile)
    else:
        sicd_meta.ImageCreation.Profile = profile
        if sicd_meta.ImageCreation.DateTime is None:
            sicd_meta.ImageCreation.DateTime = now
    return sicd_meta


def extract_clas(sicd: SICDType) -> str:
    """
    Extract the classification string from a SICD as appropriate for NITF Security
    tags CLAS attribute.

    Parameters
    ----------
    sicd : SICDType

    Returns
    -------
    str
    """
    if sicd.CollectionInfo is None or sicd.CollectionInfo.Classification is None:
        return 'U'

    c_str = sicd.CollectionInfo.Classification.upper().strip()

    if 'UNCLASS' in c_str or c_str == 'U':
        return 'U'
    elif 'CONFIDENTIAL' in c_str or c_str == 'C' or c_str.startswith('C/'):
        return 'C'
    elif 'TOP SECRET' in c_str or c_str == 'TS' or c_str.startswith('TS/'):
        return 'T'
    elif 'SECRET' in c_str or c_str == 'S' or c_str.startswith('S/'):
        return 'S'
    elif 'FOUO' in c_str.upper() or 'RESTRICTED' in c_str.upper():
        return 'R'
    else:
        logger.error(
            'Unclear how to extract CLAS for classification string {}.\n\t'
            'Should be set appropriately.'.format(c_str))
        return 'U'


def create_security_tags_from_sicd(sicd_meta: SICDType) -> NITFSecurityTags:
    def get_basic_args():
        out = {}
        sec_tags = sicd_meta.NITF.get('Security', {})
        # noinspection PyProtectedMember
        for fld in NITFSecurityTags._ordering:
            if fld in sec_tags:
                out[fld] = sec_tags[fld]
        return out

    def get_clas():
        if 'CLAS' in args:
            return
        args['CLAS'] = extract_clas(sicd_meta)

    def get_code(in_str):
        if 'CODE' in args:
            return

        # TODO: this is pretty terrible...
        code = re.search('(?<=/)[^/].*', in_str)
        if code is not None:
            args['CODE'] = code.group()

    def get_clsy():
        if args.get('CLSY', '').strip() == '':
            args['CLSY'] = 'US'

    args = get_basic_args()
    if sicd_meta.CollectionInfo is not None:
        get_clas()
        get_code(sicd_meta.CollectionInfo.Classification)
        get_clsy()

    return NITFSecurityTags(**args)


class SICDWritingDetails(NITFWritingDetails):
    """
    Manager for all the NITF subheader information associated with the SICD.

    Introduced in version 1.3.0.
    """

    __slots__ = (
        '_sicd_meta', '_security_tags', '_row_limit', '_check_older_version',
        '_required_version')

    def __init__(
            self,
            sicd_meta: SICDType,
            row_limit: Optional[int] = None,
            additional_des: Optional[Sequence[DESSubheaderManager]] = None,
            text_managers: Optional[Tuple[TextSubheaderManager, ...]] = None,
            res_managers: Optional[Tuple[RESSubheaderManager, ...]] = None,
            check_older_version: bool = False):
        """

        Parameters
        ----------
        sicd_meta : SICDType
        row_limit : None|int
            Desired row limit for the sicd image segments. Non-positive values
            or values > 99999 will be ignored.
        additional_des : None|Sequence[DESSubheaderManager]
        text_managers: Optional[Tuple[TextSubheaderManager, ...]]
        res_managers: Optional[Tuple[RESSubheaderManager, ...]]
        check_older_version : bool
            Try to create an older version sicd, for compliance
        """

        self._check_older_version = bool(check_older_version)
        self._security_tags = None
        self._sicd_meta = None
        self._set_sicd_meta(sicd_meta)
        self._required_version = self.sicd_meta.version_required()
        self._create_security_tags()
        self._row_limit = None
        self._set_row_limit(row_limit)

        header = self._create_header()
        image_managers, image_segment_collections, image_segment_coordinates = self._create_image_segments()
        des_managers = self._create_des_segments(additional_des)

        # NB: graphics not permitted in sicd
        NITFWritingDetails.__init__(
            self,
            header,
            image_managers=image_managers,
            image_segment_collections=image_segment_collections,
            image_segment_coordinates=image_segment_coordinates,
            text_managers=text_managers,
            des_managers=des_managers,
            res_managers=res_managers)

    @property
    def sicd_meta(self) -> SICDType:
        """
        SICDType: The sicd metadata
        """

        return self._sicd_meta

    def _set_sicd_meta(self, value):
        if self._sicd_meta is not None:
            raise ValueError('sicd_meta is read only')
        self._sicd_meta = validate_sicd_for_writing(value)

    @property
    def requires_version(self) -> Tuple[int, int, int]:
        """
        Tuple[int, int, int]: What is the required (at minimum) sicd version?
        """

        return self._required_version

    @property
    def row_limit(self) -> int:
        return self._row_limit

    def _set_row_limit(self, value):
        if value is not None:
            if not isinstance(value, int):
                raise TypeError('row_bounds must be an integer')
            if value < 1:
                value = None

        if value is None or value > 99999:
            value = 99999

        im_seg_limit = 10**10 - 2  # allowable image segment size
        row_memory_size = self.sicd_meta.ImageData.NumCols*self.sicd_meta.ImageData.get_pixel_size()

        memory_limit = int(numpy.floor(im_seg_limit/row_memory_size))
        self._row_limit = min(value, memory_limit)

    @property
    def security_tags(self) -> NITFSecurityTags:
        """
        NITFSecurityTags: The default NITF security tags for use.
        """

        return self._security_tags

    def _create_security_tags(self) -> None:
        """
        Creates a NITF security tags object with `CLAS` and `CODE` attributes in
        the sicd_meta.NITF property and/or extracted from the
        SICD.CollectionInfo.Classification value.

        Returns
        -------
        None
        """

        self._security_tags = create_security_tags_from_sicd(self.sicd_meta)

    def _get_ftitle(self) -> str:
        ftitle = self.sicd_meta.NITF.get('FTITLE', None)
        if ftitle is None:
            ftitle = self.sicd_meta.NITF.get('SUGGESTED_NAME', None)
        if ftitle is None:
            ftitle = self.sicd_meta.get_suggested_name(1)
        if ftitle is None and self.sicd_meta.CollectionInfo is not None and \
                self.sicd_meta.CollectionInfo.CoreName is not None:
            ftitle = 'SICD: {}'.format(self.sicd_meta.CollectionInfo.CoreName)
        if ftitle is None:
            ftitle = 'SICD: Unknown'
        if self._check_older_version and self._required_version < (1, 2, 0) and \
                not ftitle.startswith('SICD:'):
            ftitle = 'SICD:' + ftitle
        return ftitle

    def _get_fdt(self):
        return re.sub(r'[^0-9]', '', str(self.sicd_meta.ImageCreation.DateTime.astype('datetime64[s]')))

    def _get_idatim(self) -> str:
        idatim = ' '
        if self.sicd_meta.Timeline is not None and self.sicd_meta.Timeline.CollectStart is not None:
            idatim = re.sub(r'[^0-9]', '', str(self.sicd_meta.Timeline.CollectStart.astype('datetime64[s]')))
        return idatim

    def _get_ostaid(self) -> str:
        ostaid = self.sicd_meta.NITF.get('OSTAID', 'Unknown')
        return ostaid

    def _get_isorce(self) -> str:
        isorce = self.sicd_meta.NITF.get('ISORCE', None)
        if isorce is None and \
                self.sicd_meta.CollectionInfo is not None and \
                self.sicd_meta.CollectionInfo.CollectorName is not None:
            isorce = 'SICD: {}'.format(self.sicd_meta.CollectionInfo.CollectorName)
        if isorce is None:
            isorce = 'SICD: Unknown Collector'
        return isorce

    def _get_iid2(self) -> str:
        iid2 = self.sicd_meta.NITF.get('IID2', self._get_ftitle())
        if self._check_older_version and self._required_version < (1, 2, 0) and \
                not iid2.startswith('SICD:'):
            iid2 = 'SICD:' + iid2
        return iid2

    def _create_header(self) -> NITFHeader:
        """
        Create the main NITF header.

        Returns
        -------
        NITFHeader
        """

        # NB: CLEVEL and FL will be corrected...
        return NITFHeader(
            Security=self.security_tags, CLEVEL=3, OSTAID=self._get_ostaid(),
            FDT=self._get_fdt(), FTITLE=self._get_ftitle(), FL=0)

    def _create_image_segments(self) -> Tuple[
            Tuple[ImageSubheaderManager, ...],
            Tuple[Tuple[int, ...], ...],
            Tuple[Tuple[Tuple[int, ...], ...]]]:
        image_managers = []
        basic_args = {
            'IREP': 'NODISPLY',
            'IC': 'NC',
            'ICAT': 'SAR',
            'IID2': self._get_iid2(),
            'IDATIM': self._get_idatim(),
            'ISORCE': self._get_isorce()
        }
        pixel_type = self.sicd_meta.ImageData.PixelType  # required to be defined
        # NB: SICDs are required to be stored as big-endian, so the endian-ness
        #   of the memmap must be explicit
        if pixel_type == 'RE32F_IM32F':
            basic_args['PVTYPE'] = 'R'
            basic_args['NBPP'] = 32
            basic_args['ABPP'] = 32
            isubcat = ('I', 'Q')
        elif pixel_type == 'RE16I_IM16I':
            basic_args['PVTYPE'] = 'SI'
            basic_args['NBPP'] = 16
            basic_args['ABPP'] = 16
            isubcat = ('I', 'Q')
        elif pixel_type == 'AMP8I_PHS8I':
            basic_args['PVTYPE'] = 'INT'
            basic_args['NBPP'] = 8
            basic_args['ABPP'] = 8
            isubcat = ('M', 'P')
        else:
            raise ValueError('Got unhandled pixel_type {}'.format(pixel_type))

        rows = self.sicd_meta.ImageData.NumRows
        cols = self.sicd_meta.ImageData.NumCols
        icp = None
        if self.sicd_meta.GeoData is not None and self.sicd_meta.GeoData.ImageCorners is not None:
            # noinspection PyTypeChecker
            icp = self.sicd_meta.GeoData.ImageCorners.get_array(dtype=numpy.float64)

        image_segment_limits = default_image_segmentation(rows, cols, self.row_limit)

        image_segment_collection = (tuple(range(len(image_segment_limits))), )
        image_segment_coordinates = (image_segment_limits, )

        for i, entry in enumerate(image_segment_limits):
            if i == 0:
                iloc = '0000000000'
            else:
                prev_lims = image_segment_limits[i-1]
                prev_rows = prev_lims[1] - prev_lims[0]
                iloc = '{0:05d}00000'.format(prev_rows)

            this_rows = entry[1]-entry[0]
            this_cols = entry[3]-entry[2]
            subhead = ImageSegmentHeader(
                IID1='SICD{0:03d}'.format(0 if len(image_segment_limits) == 1 else i+1),
                NROWS=this_rows,
                NCOLS=this_cols,
                IGEOLO=interpolate_corner_points_string(numpy.array(entry, dtype=numpy.int64), rows, cols, icp),
                NPPBH=0 if this_cols > 8192 else this_cols,
                NPPBV=0 if this_rows > 8192 else this_rows,
                NBPC=1,
                NBPR=1,
                IDLVL=i+1,
                IALVL=i,
                ILOC=iloc,
                Bands=ImageBands(values=[ImageBand(ISUBCAT=entry) for entry in isubcat]),
                Security=self._security_tags,
                **basic_args)
            image_managers.append(ImageSubheaderManager(subhead))
        return tuple(image_managers), image_segment_collection, image_segment_coordinates

    def _create_sicd_des(self) -> DESSubheaderManager:
        uh_args = self.sicd_meta.get_des_details(self._check_older_version)
        desshdt = str(self.sicd_meta.ImageCreation.DateTime.astype('datetime64[s]'))
        if desshdt[-1] != 'Z':
            desshdt += 'Z'
        uh_args['DESSHDT'] = desshdt

        desshlpg = ''
        if self.sicd_meta.GeoData is not None and self.sicd_meta.GeoData.ImageCorners is not None:
            # noinspection PyTypeChecker
            icp = self.sicd_meta.GeoData.ImageCorners.get_array(dtype=numpy.float64)
            temp = []
            for entry in icp:
                temp.append('{0:0=+12.8f}{1:0=+13.8f}'.format(entry[0], entry[1]))
            temp.append(temp[0])
            desshlpg = ''.join(temp)
        uh_args['DESSHLPG'] = desshlpg

        subhead = DataExtensionHeader(
            Security=self.security_tags,
            UserHeader=XMLDESSubheader(**uh_args))
        return DESSubheaderManager(
            subhead, self.sicd_meta.to_xml_bytes(tag='SICD', urn=uh_args['DESSHTN']))

    def _create_des_segments(
            self,
            additional_des: Optional[Sequence[DESSubheaderManager]]) -> Tuple[DESSubheaderManager, ...]:

        if additional_des is not None:
            des_managers = list(additional_des)
        else:
            des_managers = []
        des_managers.append(self._create_sicd_des())
        return tuple(des_managers)


class SICDWriter(NITFWriter):
    """
    Writer class for a SICD file - a NITF file containing complex radar data and
    SICD data extension.

    **Changed in version 1.3.0** to reflect NITFWriter changes.
    """

    def __init__(
            self,
            file_object: Union[str, BinaryIO],
            sicd_meta: Optional[SICDType] = None,
            sicd_writing_details: Optional[SICDWritingDetails] = None,
            check_older_version: bool = False,
            check_existence: bool = True):
        """

        Parameters
        ----------
        file_object : str|BinaryIO
        sicd_meta : None|SICDType
        sicd_writing_details : None|SICDWritingDetails
        check_older_version : bool
            Try to create an older version sicd, for compliance with standard
            NGA applications like SOCET or RemoteView
        check_existence : bool
            Should we check if the given file already exists?
        """

        if sicd_meta is None and sicd_writing_details is None:
            raise ValueError('One of sicd_meta or sicd_writing_details must be provided.')
        if sicd_writing_details is None:
            sicd_writing_details = SICDWritingDetails(sicd_meta, check_older_version=check_older_version)
        NITFWriter.__init__(
            self, file_object, sicd_writing_details, check_existence=check_existence)

    @property
    def nitf_writing_details(self) -> SICDWritingDetails:
        """
        SICDWritingDetails: The SICD/NITF subheader details.
        """

        # noinspection PyTypeChecker
        return self._nitf_writing_details

    @nitf_writing_details.setter
    def nitf_writing_details(self, value):
        if self._nitf_writing_details is not None:
            raise ValueError('nitf_writing_details is read-only')
        if not isinstance(value, SICDWritingDetails):
            raise TypeError('nitf_writing_details must be of type {}'.format(SICDWritingDetails))
        self._nitf_writing_details = value

    @property
    def sicd_meta(self) -> SICDType:
        return self.nitf_writing_details.sicd_meta

    def get_format_function(
            self,
            raw_dtype: numpy.dtype,
            complex_order: Optional[str],
            lut: Optional[numpy.ndarray],
            band_dimension: int,
            image_segment_index: Optional[int] = None,
            **kwargs) -> Optional[FormatFunction]:
        if complex_order is not None and complex_order != 'IQ':
            if complex_order != 'MP' or raw_dtype.name != 'uint8' or band_dimension != 2:
                raise ValueError('Got unsupported SICD band type definition')
            if self.sicd_meta.ImageData.PixelType != 'AMP8I_PHS8I' or self.sicd_meta.ImageData.AmpTable is None:
                raise ValueError('Expected AMP8I_PHS8I')
            return AmpLookupFunction(raw_dtype, self.sicd_meta.ImageData.AmpTable)
        return NITFWriter.get_format_function(
            self, raw_dtype, complex_order, lut, band_dimension, image_segment_index, **kwargs)
